from pymethylprocess.MethylationDataTypes import MethylationArray, extract_pheno_beta_df_from_pickle_dict
from methylnet.models import *
from methylnet.datasets import get_methylation_dataset
import torch
from torch.utils.data import DataLoader
from torch.nn import MSELoss, BCELoss, CrossEntropyLoss, NLLLoss
import pickle
import pandas as pd, numpy as np
import click
import os, copy
from os.path import join
from collections import Counter

CONTEXT_SETTINGS = dict(help_option_names=['-h','--help'], max_content_width=90)

@click.group(context_settings= CONTEXT_SETTINGS)
@click.version_option(version='0.1')
def prediction():
    pass

def train_predict(train_pkl,test_pkl,input_vae_pkl,output_dir,cuda,interest_cols,categorical,disease_only,hidden_layer_topology,learning_rate_vae,learning_rate_mlp,weight_decay,dropout_p,n_epochs, scheduler='null', decay=0.5, t_max=10, eta_min=1e-6, t_mult=2, batch_size=50, val_pkl='val_methyl_array.pkl', n_workers=8, add_validation_set=False, loss_reduction='sum', add_softmax=False):
    os.makedirs(output_dir,exist_ok=True)

    output_file = join(output_dir,'results.csv')
    training_curve_file = join(output_dir, 'training_val_curve.p')
    results_file = join(output_dir,'results.p')
    output_file_latent = join(output_dir,'latent.csv')
    output_model = join(output_dir,'output_model.p')
    output_pkl = join(output_dir, 'vae_mlp_methyl_arr.pkl')
    output_onehot_encoder = join(output_dir, 'one_hot_encoder.p')

    #input_dict = pickle.load(open(input_pkl,'rb'))
    if cuda:
        vae_model = torch.load(input_vae_pkl)
        vae_model.cuda_on=True
    else:
        vae_model = torch.load(input_vae_pkl,map_location='cpu')
        vae_model.cuda_on=False

    train_methyl_array, val_methyl_array, test_methyl_array = MethylationArray.from_pickle(train_pkl), MethylationArray.from_pickle(val_pkl), MethylationArray.from_pickle(test_pkl)#methyl_array.split_train_test(train_p=train_percent, stratified=(True if categorical else False), disease_only=disease_only, key=interest_cols[0], subtype_delimiter=',')

    if not categorical:
        train_methyl_array.remove_na_samples(interest_cols)
        val_methyl_array.remove_na_samples(interest_cols)
        test_methyl_array.remove_na_samples(interest_cols)

    print(train_methyl_array.beta.shape)
    print(val_methyl_array.beta.shape)
    print(test_methyl_array.beta.shape)

    if len(interest_cols) == 1 and disease_only and interest_cols[0].endswith('_only')==False:
        print(interest_cols)
        interest_cols[0] += '_only'
        print(train_methyl_array.pheno[interest_cols[0]].unique())
        print(test_methyl_array.pheno[interest_cols[0]].unique())

    train_methyl_dataset = get_methylation_dataset(train_methyl_array,interest_cols,categorical=categorical, predict=True) # train, test split? Add val set?
    #print(list(train_methyl_dataset.encoder.get_feature_names()))
    val_methyl_dataset = get_methylation_dataset(val_methyl_array,interest_cols,categorical=categorical, predict=True, categorical_encoder=train_methyl_dataset.encoder)
    test_methyl_dataset = get_methylation_dataset(test_methyl_array,interest_cols,categorical=categorical, predict=True, categorical_encoder=train_methyl_dataset.encoder)


    if not batch_size:
        batch_size=len(train_methyl_dataset)
    train_batch_size = min(batch_size,len(train_methyl_dataset))
    val_batch_size = min(batch_size,len(val_methyl_dataset))

    train_methyl_dataloader = DataLoader(
        dataset=train_methyl_dataset,
        num_workers=n_workers,
        batch_size=train_batch_size,
        shuffle=True)

    val_methyl_dataloader = DataLoader(
        dataset=val_methyl_dataset,
        num_workers=n_workers,
        batch_size=val_batch_size,
        shuffle=True) # False

    test_methyl_dataloader = DataLoader(
        dataset=test_methyl_dataset,
        num_workers=n_workers,
        batch_size=min(batch_size,len(test_methyl_dataset)),
        shuffle=False)

    scaling_factors = dict(val=float(len(val_methyl_dataset))/((len(val_methyl_dataset)//val_batch_size)*val_batch_size),
                           train_batch_size=train_batch_size,val_batch_size=val_batch_size)

    model=VAE_MLP(vae_model=vae_model,categorical=categorical,hidden_layer_topology=hidden_layer_topology,n_output=train_methyl_dataset.outcome_col.shape[1],dropout_p=dropout_p, add_softmax=add_softmax)

    class_weights=[]
    if categorical:
        out_weight=Counter(np.argmax(train_methyl_dataset.outcome_col,axis=1))
        #total_samples=sum(out_weight.values())
        for k in sorted(list(out_weight.keys())):
            class_weights.append(1./float(out_weight[k])) # total_samples
        class_weights=np.array(class_weights)
        class_weights=(class_weights/class_weights.sum()).tolist()
        print(class_weights)

    if class_weights:
        class_weights = torch.FloatTensor(class_weights)
        if cuda:
            class_weights = class_weights.cuda()
    else:
        class_weights = None

    optimizer_vae = torch.optim.Adam(model.vae.parameters(), lr = learning_rate_vae, weight_decay=weight_decay)
    optimizer_mlp = torch.optim.Adam(model.mlp.parameters(), lr = learning_rate_mlp, weight_decay=weight_decay)
    loss_fn = CrossEntropyLoss(reduction=loss_reduction,weight= class_weights) if categorical else MSELoss(reduction=loss_reduction) # 'sum'
    scheduler_opts=dict(scheduler=scheduler,lr_scheduler_decay=decay,T_max=t_max,eta_min=eta_min,T_mult=t_mult)
    vae_mlp=MLPFinetuneVAE(mlp_model=model,n_epochs=n_epochs,categorical=categorical,loss_fn=loss_fn,optimizer_vae=optimizer_vae,optimizer_mlp=optimizer_mlp,cuda=cuda, scheduler_opts=scheduler_opts)
    if add_validation_set:
        vae_mlp.add_validation_set(val_methyl_dataloader)
    vae_mlp = vae_mlp.fit(train_methyl_dataloader)
    if 'encoder' in dir(train_methyl_dataset):
        pickle.dump(train_methyl_dataset.encoder,open(output_onehot_encoder,'wb'))
    results = dict(test={},train={},val={})
    results['train']['y_pred'], results['train']['y_true'], _, _ = vae_mlp.predict(train_methyl_dataloader)
    results['val']['y_pred'], results['val']['y_true'], _, _ = vae_mlp.predict(val_methyl_dataloader)
    del train_methyl_dataloader, train_methyl_dataset
    """methyl_dataset=get_methylation_dataset(methyl_array,interest_cols,predict=True)
    methyl_dataset_loader = DataLoader(
        dataset=methyl_dataset,
        num_workers=9,
        batch_size=1,
        shuffle=False)"""
    Y_pred, Y_true, latent_projection, sample_names = vae_mlp.predict(test_methyl_dataloader) # FIXME change to include predictions for all classes for AUC
    results['test']['y_pred'], results['test']['y_true'] = copy.deepcopy(Y_pred), copy.deepcopy(Y_true)
    if categorical:
        Y_true=Y_true.argmax(axis=1)[:,np.newaxis]
        Y_pred=Y_pred.argmax(axis=1)[:,np.newaxis]
    test_methyl_array = test_methyl_dataset.to_methyl_array()
    """if categorical:
        Y_true=test_methyl_dataset.encoder.inverse_transform(Y_true)[:,np.newaxis]
        Y_pred=test_methyl_dataset.encoder.inverse_transform(Y_pred)[:,np.newaxis]"""
    #sample_names = np.array(list(test_methyl_array.beta.index)) # FIXME
    #outcomes = np.array([outcome[0] for outcome in outcomes]) # FIXME
    Y_pred=pd.DataFrame(Y_pred.flatten() if (np.array(Y_pred.shape)==1).any() else Y_pred,index=test_methyl_array.beta.index,columns=(['y_pred'] if categorical else interest_cols))#dict(zip(sample_names,outcomes))
    Y_true=pd.DataFrame(Y_true.flatten() if (np.array(Y_true.shape)==1).any() else Y_true,index=test_methyl_array.beta.index,columns=(['y_true'] if categorical else interest_cols))
    results_df = pd.concat([Y_pred,Y_true],axis=1) if categorical else pd.concat([Y_pred.rename(columns={name:name+'_pred' for name in list(Y_pred)}),Y_true.rename(columns={name:name+'_true' for name in list(Y_pred)})],axis=1)  # FIXME
    latent_projection=pd.DataFrame(latent_projection,index=test_methyl_array.beta.index)
    test_methyl_array.beta=latent_projection
    test_methyl_array.write_pickle(output_pkl)
    pickle.dump(results,open(results_file,'wb'))
    pickle.dump(vae_mlp.training_plot_data,open(training_curve_file,'wb'))
    latent_projection.to_csv(output_file_latent)
    torch.save(vae_mlp.model,output_model)
    results_df.to_csv(output_file)#pickle.dump(outcome_dict, open(outcome_dict_file,'wb'))
    return latent_projection, Y_pred, Y_true, vae_mlp, scaling_factors

# ADD OUTPUT METRICS AND TRAINING PLOT CURVE

@prediction.command() # FIXME finish this!!
@click.option('-i', '--train_pkl', default='./train_val_test_sets/train_methyl_array.pkl', help='Input database for beta and phenotype data.', type=click.Path(exists=False), show_default=True)
@click.option('-tp', '--test_pkl', default='./train_val_test_sets/test_methyl_array.pkl', help='Test database for beta and phenotype data.', type=click.Path(exists=False), show_default=True)
@click.option('-vae', '--input_vae_pkl', default='./embeddings/output_model.p', help='Trained VAE.', type=click.Path(exists=False), show_default=True)
@click.option('-o', '--output_dir', default='./predictions/', help='Output directory for predictions.', type=click.Path(exists=False), show_default=True)
@click.option('-c', '--cuda', is_flag=True, help='Use GPUs.')
@click.option('-ic', '--interest_cols', default=['disease'], multiple=True, help='Specify columns looking to make predictions on.', show_default=True)
@click.option('-cat', '--categorical', is_flag=True, help='Multi-class prediction.', show_default=True)
@click.option('-do', '--disease_only', is_flag=True, help='Only look at disease, or text before subtype_delimiter.')
@click.option('-hlt', '--hidden_layer_topology', default='', help='Topology of hidden layers, comma delimited, leave empty for one layer encoder, eg. 100,100 is example of 5-hidden layer topology.', type=click.Path(exists=False), show_default=True)
@click.option('-lr_vae', '--learning_rate_vae', default=1e-5, help='Learning rate VAE.', show_default=True)
@click.option('-lr_mlp', '--learning_rate_mlp', default=1e-3, help='Learning rate MLP.', show_default=True)
@click.option('-wd', '--weight_decay', default=1e-4, help='Weight decay of adam optimizer.', show_default=True)
@click.option('-dp', '--dropout_p', default=0.2, help='Dropout Percentage.', show_default=True)
@click.option('-e', '--n_epochs', default=50, help='Number of epochs to train over.', show_default=True)
@click.option('-s', '--scheduler', default='null', help='Type of learning rate scheduler.', type=click.Choice(['null','exp','warm_restarts']),show_default=True)
@click.option('-d', '--decay', default=0.5, help='Learning rate scheduler decay for exp selection.', show_default=True)
@click.option('-t', '--t_max', default=10, help='Number of epochs before cosine learning rate restart.', show_default=True)
@click.option('-eta', '--eta_min', default=1e-6, help='Minimum cosine LR.', show_default=True)
@click.option('-m', '--t_mult', default=2., help='Multiply current restart period times this number given number of restarts.', show_default=True)
@click.option('-bs', '--batch_size', default=50, show_default=True, help='Batch size.')
@click.option('-vp', '--val_pkl', default='./train_val_test_sets/val_methyl_array.pkl', help='Validation Set Methylation Array Location.', show_default=True, type=click.Path(exists=False),)
@click.option('-w', '--n_workers', default=9, show_default=True, help='Number of workers.')
@click.option('-v', '--add_validation_set', is_flag=True, help='Evaluate validation set.')
@click.option('-l', '--loss_reduction', default='sum', show_default=True, help='Type of reduction on loss function.', type=click.Choice(['sum','elementwise_mean','none']))
@click.option('-hl', '--hyperparameter_log', default='predictions/predict_hyperparameters_log.csv', show_default=True, help='CSV file containing prior runs.', type=click.Path(exists=False))
@click.option('-j', '--job_name', default='predict_job', show_default=True, help='Embedding job name.', type=click.Path(exists=False))
@click.option('-sft', '--add_softmax', is_flag=True, help='Add softmax for predicting probability distributions. Experimental.')
def make_prediction(train_pkl,test_pkl,input_vae_pkl,output_dir,cuda,interest_cols,categorical,disease_only,hidden_layer_topology,learning_rate_vae,learning_rate_mlp,weight_decay,dropout_p,n_epochs, scheduler='null', decay=0.5, t_max=10, eta_min=1e-6, t_mult=2, batch_size=50, val_pkl='val_methyl_array.pkl', n_workers=8, add_validation_set=False, loss_reduction='sum', hyperparameter_log='predictions/predict_hyperparameters_log.csv', job_name='predict_job', add_softmax=False):
    """Train prediction model by fine-tuning VAE and appending/training MLP to make classification/regression predictions on MethylationArrays."""
    hlt_list=filter(None,hidden_layer_topology.split(','))
    if hlt_list:
        hidden_layer_topology=list(map(int,hlt_list))
    else:
        hidden_layer_topology=[]
    latent_projection, Y_pred, Y_true, vae_mlp, scaling_factors = train_predict(train_pkl,test_pkl,input_vae_pkl,output_dir,cuda,list(interest_cols),categorical,disease_only,hidden_layer_topology,learning_rate_vae,learning_rate_mlp,weight_decay,dropout_p,n_epochs, scheduler, decay, t_max, eta_min, t_mult, batch_size, val_pkl, n_workers, add_validation_set, loss_reduction, add_softmax)
    accuracy, precision, recall, f1 = -1,-1,-1,-1
    if categorical:
        from sklearn.metrics import accuracy_score, precision_score, recall_score, f1_score
        accuracy, precision, recall, f1 = accuracy_score(Y_true,Y_pred), precision_score(Y_true,Y_pred,average='weighted'), recall_score(Y_true,Y_pred,average='weighted'), f1_score(Y_true,Y_pred,average='weighted')
    else:
        from sklearn.metrics import mean_squared_error, r2_score, explained_variance_score, mean_absolute_error
        accuracy, precision, recall, f1 = mean_squared_error(Y_true,Y_pred,multioutput='raw_values'), r2_score(Y_true,Y_pred,multioutput='raw_values'), explained_variance_score(Y_true,Y_pred,multioutput='raw_values'), mean_absolute_error(Y_true,Y_pred,multioutput='raw_values')
        if len(interest_cols)>1:
            accuracy, precision, recall, f1 = ';'.join([str(v) for v in accuracy]),';'.join([str(v) for v in precision]),';'.join([str(v) for v in recall]),';'.join([str(v) for v in f1])
        else:
            accuracy, precision, recall, f1=accuracy[0], precision[0], recall[0], f1[0]
    hyperparameter_row = [job_name,n_epochs, vae_mlp.best_epoch, vae_mlp.min_loss, vae_mlp.min_val_loss, vae_mlp.min_val_loss*scaling_factors['val'], accuracy, precision, recall, f1, vae_mlp.model.vae.n_input, vae_mlp.model.vae.n_latent, str(hidden_layer_topology), learning_rate_vae, learning_rate_mlp, weight_decay, scheduler, t_max, t_mult, eta_min, scaling_factors['train_batch_size'], scaling_factors['val_batch_size'], dropout_p]
    hyperparameter_df = pd.DataFrame(columns=['job_name','n_epochs',"best_epoch", "min_loss", "min_val_loss", "min_val_loss-batchsize_adj","test_accuracy" if categorical else 'neg_mean_squared_error', "test_precision" if categorical else 'r2_score', "test_recall" if categorical else 'explained_variance', "test_f1" if categorical else 'mean_absolute_error', "n_input", "n_latent", "hidden_layer_encoder_topology", "learning_rate_vae", "learning_rate_mlp", "weight_decay", "scheduler", "t_max", "t_mult", "eta_min","train_batch_size", "val_batch_size", "dropout_p"])
    hyperparameter_df.loc[0] = hyperparameter_row
    if os.path.exists(hyperparameter_log):
        print('APPEND')
        hyperparameter_df_former = pd.read_csv(hyperparameter_log)
        hyperparameter_df_former=hyperparameter_df_former[[col for col in list(hyperparameter_df) if not col.startswith('Unnamed')]]
        hyperparameter_df=pd.concat([hyperparameter_df_former,hyperparameter_df],axis=0)
    hyperparameter_df.to_csv(hyperparameter_log)

@prediction.command()
@click.option('-tp', '--test_pkl', default='./train_val_test_sets/test_methyl_array.pkl', help='Test database for beta and phenotype data.', type=click.Path(exists=False), show_default=True)
@click.option('-m', '--model_pickle', default='./predictions/output_model.p', help='Pytorch model containing forward_predict method.', type=click.Path(exists=False), show_default=True)
@click.option('-bs', '--batch_size', default=50, show_default=True, help='Batch size.')
@click.option('-w', '--n_workers', default=9, show_default=True, help='Number of workers.')
@click.option('-ic', '--interest_cols', default=['disease'], multiple=True, help='Specify columns looking to make predictions on.', show_default=True)
@click.option('-cat', '--categorical', is_flag=True, help='Multi-class prediction.', show_default=True)
@click.option('-c', '--cuda', is_flag=True, help='Use GPUs.')
@click.option('-e', '--categorical_encoder', default='./predictions/one_hot_encoder.p', help='One hot encoder if categorical model. If path exists, then return top positive controbutions per samples of that class. Encoded values must be of sample class as interest_col.', type=click.Path(exists=False), show_default=True)
@click.option('-o', '--output_dir', default='./new_predictions/', help='Output directory for predictions.', type=click.Path(exists=False), show_default=True)
def make_new_predictions(test_pkl, model_pickle, batch_size, n_workers, interest_cols, categorical, cuda, categorical_encoder, output_dir):
    """Run prediction model again to further assess outcome. Only evaluate prediction model."""
    os.makedirs(output_dir,exist_ok=True)
    test_methyl_array = MethylationArray.from_pickle(test_pkl) # generate results pickle to run through classification/regression report
    if cuda:
        model = torch.load(model_pickle)
        model.vae.cuda_on=True
    else:
        model = torch.load(model_pickle,map_location='cpu')
        model.vae.cuda_on=False
    if not categorical:
        test_methyl_array.remove_na_samples(interest_cols if len(interest_cols)>1 else interest_cols[0])
    if os.path.exists(categorical_encoder):
        categorical_encoder=pickle.load(open(categorical_encoder,'rb'))
    else:
        categorical_encoder=None
    test_methyl_dataset = get_methylation_dataset(test_methyl_array,interest_cols,categorical=categorical, predict=True, categorical_encoder=categorical_encoder)
    test_methyl_dataloader = DataLoader(
        dataset=test_methyl_dataset,
        num_workers=n_workers,
        batch_size=min(batch_size,len(test_methyl_dataset)),
        shuffle=False)
    vae_mlp=MLPFinetuneVAE(mlp_model=model,categorical=categorical,cuda=cuda)

    Y_pred, Y_true, latent_projection, sample_names = vae_mlp.predict(test_methyl_dataloader)

    results = dict(test={})
    results['test']['y_pred'], results['test']['y_true'] = copy.deepcopy(Y_pred), copy.deepcopy(Y_true)

    if categorical:
        Y_true=Y_true.argmax(axis=1)[:,np.newaxis]
        Y_pred=Y_pred.argmax(axis=1)[:,np.newaxis]
    test_methyl_array = test_methyl_dataset.to_methyl_array()

    Y_pred=pd.DataFrame(Y_pred.flatten() if (np.array(Y_pred.shape)==1).any() else Y_pred,index=test_methyl_array.beta.index,columns=(['y_pred'] if categorical else interest_cols))
    Y_true=pd.DataFrame(Y_true.flatten() if (np.array(Y_true.shape)==1).any() else Y_true,index=test_methyl_array.beta.index,columns=(['y_true'] if categorical else interest_cols))
    results_df = pd.concat([Y_pred,Y_true],axis=1) if categorical else pd.concat([Y_pred.rename(columns={name:name+'_pred' for name in list(Y_pred)}),Y_true.rename(columns={name:name+'_true' for name in list(Y_pred)})],axis=1)  # FIXME
    latent_projection=pd.DataFrame(latent_projection,index=test_methyl_array.beta.index)
    test_methyl_array.beta=latent_projection

    output_file = join(output_dir,'results.csv')
    results_file = join(output_dir,'results.p')
    output_file_latent = join(output_dir,'latent.csv')
    output_pkl = join(output_dir, 'vae_mlp_methyl_arr.pkl')

    test_methyl_array.write_pickle(output_pkl)
    pickle.dump(results,open(results_file,'wb'))
    latent_projection.to_csv(output_file_latent)
    results_df.to_csv(output_file)

@prediction.command()
@click.option('-hcsv', '--hyperparameter_input_csv', default='predictions/predict_hyperparameters_scan_input.csv', show_default=True, help='CSV file containing hyperparameter inputs.', type=click.Path(exists=False))
@click.option('-hl', '--hyperparameter_output_log', default='predictions/predict_hyperparameters_log.csv', show_default=True, help='CSV file containing prior runs.', type=click.Path(exists=False))
@click.option('-g', '--generate_input', is_flag=True, help='Generate hyperparameter input csv.')
@click.option('-c', '--job_chunk_size', default=4, help='If not series, chunk up and run these number of commands at once..')
@click.option('-ic', '--interest_cols', default=['disease_only'], multiple=True, help='Column to stratify samples on.')
@click.option('-cat', '--categorical', is_flag=True, help='Whether to run categorical analysis or not.')
@click.option('-r', '--reset_all', is_flag=True, help='Run all jobs again.')
@click.option('-t', '--torque', is_flag=True, help='Submit jobs on torque.')
@click.option('-gpu', '--gpu', default=-1, help='If torque submit, which gpu to use.', show_default=True)
@click.option('-gn', '--gpu_node', default=-1, help='If torque submit, which gpu node to use.', show_default=True)
@click.option('-nh', '--nohup', is_flag=True, help='Nohup launch jobs.')
@click.option('-n', '--n_jobs_relaunch', default=0, help='Relaunch n top jobs from previous run.', show_default=True)
@click.option('-cp', '--crossover_p', default=0., help='Rate of crossover between hyperparameters.', show_default=True)
@click.option('-mc', '--model_complexity_factor', default=1., help='Degree of neural network model complexity for hyperparameter search. Search for less wide networks with a lower complexity value, bounded between 0 and infinity.', show_default=True)
@click.option('-j', '--n_jobs', default=4, help='Number of jobs to generate.')
@click.option('-v', '--val_loss_column', default="min_val_loss-batchsize_adj", help='Validation loss column.', type=click.Path(exists=False))
@click.option('-sft', '--add_softmax', is_flag=True, help='Add softmax for predicting probability distributions.')
@click.option('-a', '--additional_command', default='', help='Additional command to input for torque run.', type=click.Path(exists=False))
@click.option('-cu', '--cuda', is_flag=True, help='Use GPUs.')
@click.option('-grid', '--hyperparameter_yaml', default='', help='YAML file with custom subset of hyperparameter grid.', type=click.Path(exists=False))
def launch_hyperparameter_scan(hyperparameter_input_csv, hyperparameter_output_log, generate_input, job_chunk_size, interest_cols, categorical, reset_all, torque, gpu, gpu_node, nohup, n_jobs_relaunch, crossover_p, model_complexity_factor,n_jobs, val_loss_column, add_softmax, additional_command, cuda, hyperparameter_yaml):
    """Run randomized hyperparameter scan of neural network hyperparameters."""
    from methylnet.hyperparameter_scans import coarse_scan, find_top_jobs
    custom_jobs=[]
    if n_jobs_relaunch:
        custom_jobs=find_top_jobs(hyperparameter_input_csv, hyperparameter_output_log,n_jobs_relaunch, crossover_p, val_loss_column)
    if os.path.exists(hyperparameter_yaml):
        from ruamel.yaml import safe_load as load
        with open(hyperparameter_yaml) as f:
            new_grid = load(f.read())
    else:
        new_grid = {}
    coarse_scan(hyperparameter_input_csv, hyperparameter_output_log, generate_input, job_chunk_size, interest_cols, reset_all, torque, gpu, gpu_node, nohup, mlp=True, custom_jobs=custom_jobs, model_complexity_factor=model_complexity_factor,n_jobs=n_jobs, categorical=categorical,add_softmax=add_softmax, additional_command=additional_command, cuda=cuda, new_grid=new_grid)

@prediction.command()
@click.option('-r', '--results_pickle', default='predictions/results.p', show_default=True, help='Results from training, validation, and testing.', type=click.Path(exists=False))
@click.option('-o', '--output_dir', default='results/', show_default=True, help='Output directory.', type=click.Path(exists=False))
def regression_report(results_pickle,output_dir):
    """Generate regression report that gives concise results from regression tasks."""
    # FIXME expand functionality
    import matplotlib
    matplotlib.use('Agg')
    import matplotlib.pyplot as plt
    import seaborn as sns
    sns.set(style="whitegrid")
    os.makedirs(output_dir,exist_ok=True)
    results_dict=pickle.load(open(results_pickle,'rb'))
    for k in results_dict:
        df_pred=pd.DataFrame(results_dict[k]['y_pred']).melt()
        df_pred=df_pred.rename(columns=dict(value='y_pred',variable='outcome_vars'))
        df_true=pd.DataFrame(results_dict[k]['y_true']).melt()
        df_true=df_true.rename(columns=dict(value='y_true',variable='true_vars'))
        df=pd.concat([df_pred,df_true],axis=1)
        #print(df)
        df=df[['outcome_vars','y_pred','y_true']]
        plt.figure()
        sns.lmplot(x="y_pred", y="y_true", col="outcome_vars", hue="outcome_vars",data=df, col_wrap=2)
        plt.savefig(os.path.join(output_dir,'{}_regression_results.png'.format(k)),dpi=300)

@prediction.command()
@click.option('-r', '--results_pickle', default='predictions/results.p', show_default=True, help='Results from training, validation, and testing.', type=click.Path(exists=False))
@click.option('-o', '--output_dir', default='results/', show_default=True, help='Output directory.', type=click.Path(exists=False))
@click.option('-e', '--categorical_encoder', default='./predictions/one_hot_encoder.p', help='One hot encoder if categorical model. If path exists, then return top positive controbutions per samples of that class. Encoded values must be of sample class as interest_col.', type=click.Path(exists=False), show_default=True)
@click.option('-a', '--average_mechanism', default='micro', show_default=True, help='Output directory.', type=click.Choice(['weighted','macro','micro','binary','samples']))
def classification_report(results_pickle,output_dir, categorical_encoder, average_mechanism):
    """Generate classification report that gives results from classification tasks."""
    from mlxtend.evaluate import bootstrap
    from sklearn.metrics import accuracy_score, recall_score, precision_score, f1_score, roc_curve, roc_auc_score, confusion_matrix
    os.makedirs(output_dir,exist_ok=True)

    def extract_ys(Y):
        return Y[:,0], Y[:,1:]

    def accuracy(Y):
        y_true, y_pred=extract_ys(Y)
        y_pred = np.argmax(y_pred,axis=1)
        return accuracy_score(y_true, y_pred)

    def recall(Y):
        y_true, y_pred=extract_ys(Y)
        y_pred = np.argmax(y_pred,axis=1)
        return recall_score(y_true, y_pred, average=average_mechanism)

    def precision(Y):
        y_true, y_pred=extract_ys(Y)
        y_pred = np.argmax(y_pred,axis=1)
        return precision_score(y_true, y_pred, average=average_mechanism)

    def f1(Y):
        y_true, y_pred=extract_ys(Y)
        y_pred = np.argmax(y_pred,axis=1)
        return f1_score(y_true, y_pred, average=average_mechanism)

    def auc(Y):
        y_true, y_pred=extract_ys(Y)
        y_pred_labels=np.argmax(y_pred,1)
        supports={i:sum((y_pred_labels[np.squeeze(y_true==i)]==i).astype(int)) for i in np.unique(y_true)}
        final_auc_score=0.
        for i in supports:
            final_auc_score+=supports[i]*roc_auc_score((y_true==i).astype(int), y_pred[:,int(i)], average='weighted')
        final_auc_score/=sum(list(supports.values()))
        return final_auc_score

    def to_y_probas(y_pred):
        return np.exp(y_pred)/np.exp(y_pred).sum(1)[:,np.newaxis]

    results_dict=pickle.load(open(results_pickle,'rb'))
    if os.path.exists(categorical_encoder):
        categorical_encoder=pickle.load(open(categorical_encoder,'rb'))
    else:
        categorical_encoder=None
    df_roc=[]
    final_results=[]
    for k in results_dict:
        y_true=results_dict[k]['y_true']
        y_true_labels=np.argmax(y_true,axis=1)[:,np.newaxis]
        classes=np.unique(y_true_labels)
        y_pred=to_y_probas(results_dict[k]['y_pred'])
        y_pred_labels=np.argmax(y_pred,1).reshape((-1,1))
        #print(y_true_labels, y_pred_labels)
        out_classes = classes.astype(int)# if categorical_encoder == None else categorical_encoder.inverse_transform(classes.astype(int).reshape((-1,1)))
        class_labels = out_classes if categorical_encoder == None else categorical_encoder.inverse_transform(out_classes)
        pd.DataFrame(confusion_matrix(y_true_labels.astype(int).flatten(),
                                      y_pred_labels.astype(int).flatten(),labels=out_classes),index=class_labels,columns=class_labels).to_csv(join(output_dir,'{}_confusion_mat.csv'.format(k)))
        Y=np.hstack((y_true_labels,y_pred))
        supports={i:sum((y_pred_labels[np.squeeze(y_true_labels==i)]==i).astype(int)) for i in classes}
        fpr = dict()
        tpr = dict()
        for i in supports:
            fpr[i], tpr[i], _ = roc_curve(y_true[:,i], y_pred[:,i])
        all_fpr = np.unique(np.concatenate([fpr[i] for i in supports]))
        mean_tpr = np.zeros_like(all_fpr)
        for i in supports:
            mean_tpr += supports[i]*np.interp(all_fpr, fpr[i], tpr[i])
        mean_tpr/=sum(list(supports.values()))
        tpr,fpr=mean_tpr,all_fpr
        df=pd.DataFrame({'fpr' : fpr, 'tpr' : tpr})
        df['Legend'] = "{} Weighted ROC, AUC={}".format(k,round(auc(Y),2))
        df_roc.append(df)
        fns = dict(accuracy=accuracy,recall=recall,precision=precision,f1=f1,auc=auc)
        for fn in fns:
            print(k,fn)
            original, std_err, ci_bounds = bootstrap(Y, num_rounds=1000,
                                             func=fns[fn],
                                             ci=0.95,
                                             seed=123)
            low,high=ci_bounds
            final_results.append([k,fn,original,std_err,low,high])
    final_results=pd.DataFrame(final_results,columns=['DataSet','Metric','Score','Error','95% CI Low','95% CI High'])
    df_roc=pd.concat(df_roc)
    final_results.to_csv(join(output_dir,'final_classification_results.csv'))
    df_roc.to_csv(join(output_dir,'Weighted_ROC.csv'))

#################

if __name__ == '__main__':
    prediction()



import pandas as pd
import numpy as np
import networkx as nx
import click
import pickle
from sklearn.preprocessing import LabelEncoder

CONTEXT_SETTINGS = dict(help_option_names=['-h','--help'], max_content_width=90)

@click.group(context_settings= CONTEXT_SETTINGS)
@click.version_option(version='0.1')
def visualize():
    pass

@visualize.command()
@click.option('-t', '--training_curve_file', default='predictions/training_val_curve.p', show_default=True, help='Training and validation loss and learning curves.', type=click.Path(exists=False))
@click.option('-o', '--outputfilename', default='results/training_curve.png', show_default=True, help='Output image.', type=click.Path(exists=False))
@click.option('-vae', '--vae_train', is_flag=True, help='Plot VAE Training curves.')
@click.option('-thr', '--threshold', default=1e10, help='Loss values get rid of greater than this.')
def plot_training_curve(training_curve_file, outputfilename, vae_train, threshold):
    """Plot training curves as output from either the VAE training or VAE MLP."""
    import os
    os.makedirs(outputfilename[:outputfilename.rfind('/')],exist_ok=True)
    import matplotlib
    matplotlib.use('Agg')
    import matplotlib.pyplot as plt
    import seaborn as sns
    #from scipy.signal import find_peaks_cwt
    #from impyute import moving_window
    def impute_peaks(signal):
        signal=signal.values
        idxs=signal>=threshold
        signal[idxs]=np.nan
        """for idx in idxs:
            signal[idx]=(signal[idx-1]+signal[idx+1])/2 if idx else signal[idx+1]"""
        return signal
    def impute_df(df):
        df=df.apply(impute_peaks,axis=0).fillna(method='bfill',axis=0)
        print(df)
        return df
    sns.set(style="whitegrid")
    df=pd.DataFrame(pickle.load(open(training_curve_file,'rb')))
    plt.subplots(1,3 if vae_train else 2,figsize=(15 if vae_train else 10,5))
    plt.subplot(131 if vae_train else 121)
    if vae_train:
        df['loss']=df['recon_loss']+df['kl_loss']
        df['val_loss']=df['val_recon_loss']+df['val_kl_loss']
    sns.lineplot(data=impute_df(df[['loss','val_loss']+(['recon_loss','val_recon_loss'] if vae_train else [])]), palette="tab10", linewidth=2.5)
    plt.xlabel('Epoch')
    plt.ylabel('Loss')
    plt.tight_layout()
    plt.subplot(132 if vae_train else 122)
    sns.lineplot(data=df[['lr_vae', 'lr_mlp'] if not vae_train else ['lr']], palette="tab10", linewidth=2.5)
    plt.xlabel('Epoch')
    plt.ylabel('Learning Rate')
    if vae_train:
        plt.subplot(133)
        sns.lineplot(data=impute_df(df[['kl_loss','val_kl_loss']]), palette="tab10", linewidth=2.5)
        plt.xlabel('Epoch')
        plt.ylabel('Loss')
    plt.tight_layout()
    plt.savefig(outputfilename)

@visualize.command()
@click.option('-r', '--roc_curve_csv', default='results/Weighted_ROC.csv', show_default=True, help='Weighted ROC Curve.', type=click.Path(exists=False))
@click.option('-o', '--outputfilename', default='results/roc_curve.png', show_default=True, help='Output image.', type=click.Path(exists=False))
def plot_roc_curve(roc_curve_csv, outputfilename):
    """Plot ROC Curves from classification tasks; requires classification_report be run first."""
    import os
    from rpy2.robjects.packages import importr
    import rpy2.robjects as robjects
    os.makedirs(outputfilename[:outputfilename.rfind('/')],exist_ok=True)
    tidyverse=importr('tidyverse')
    robjects.r("""function (in.csv, out.file.name) {
        df<-read.csv(in.csv)
        df %>%
            ggplot() +
            geom_line(aes(x = fpr, y = tpr,color=Legend)) +
            ggtitle('ROC Curve') +
            xlab('1-Specificity') +
            ylab('Sensitivity')
        ggsave(out.file.name)
        }""")(roc_curve_csv,outputfilename)

#################

if __name__ == '__main__':
    visualize()



#https://www.bioconductor.org/packages/devel/bioc/vignettes/missMethyl/inst/doc/missMethyl.html#gene-ontology-analysis
import shap, numpy as np, pandas as pd
import torch
from methylnet.datasets import RawBetaArrayDataSet, Transformer
from torch.autograd import Variable
from torch.utils.data import DataLoader
from os.path import join
import click
import os, copy
import pickle
from pymethylprocess.MethylationDataTypes import MethylationArray, MethylationArrays, extract_pheno_beta_df_from_pickle_dict
from methylnet.interpretation_classes import *


# after attaching classifier to VAE

# use SHAP to extract CpGs

# use gometh on extracted CpGs for each class in classifier!

# output to file and analyze


CONTEXT_SETTINGS = dict(help_option_names=['-h','--help'], max_content_width=90)

@click.group(context_settings= CONTEXT_SETTINGS)
@click.version_option(version='0.1')
def interpret():
    pass

@interpret.command() # FIXME add abs or just -1, positive contributions or any contributions
@click.option('-i', '--train_pkl', default='./train_val_test_sets/train_methyl_array.pkl', help='Input database for beta and phenotype data. Use ./predictions/vae_mlp_methyl_arr.pkl or ./embeddings/vae_methyl_arr.pkl for vae interpretations.', type=click.Path(exists=False), show_default=True)
@click.option('-v', '--val_pkl', default='./train_val_test_sets/val_methyl_array.pkl', help='Val database for beta and phenotype data. Use ./predictions/vae_mlp_methyl_arr.pkl or ./embeddings/vae_methyl_arr.pkl for vae interpretations.', type=click.Path(exists=False), show_default=True)
@click.option('-t', '--test_pkl', default='./train_val_test_sets/test_methyl_array.pkl', help='Pickle containing testing set.', type=click.Path(exists=False), show_default=True)
@click.option('-m', '--model_pickle', default='./predictions/output_model.p', help='Pytorch model containing forward_predict method.', type=click.Path(exists=False), show_default=True)
@click.option('-w', '--n_workers', default=9, show_default=True, help='Number of workers.')
@click.option('-bs', '--batch_size', default=512, show_default=True, help='Batch size.')
@click.option('-c', '--cuda', is_flag=True, help='Use GPUs.')
@click.option('-ns', '--n_samples', default=200, show_default=True, help='Number of samples for SHAP output.')
@click.option('-nf', '--n_top_features', default=20, show_default=True, help='Top features to select for shap outputs. If feature selection, instead use this number to choose number features that make up top global score.')
@click.option('-o', '--output_dir', default='./interpretations/shapley_explanations/', help='Output directory for interpretations.', type=click.Path(exists=False), show_default=True)
@click.option('-mth', '--method', default='kernel', help='Explainer type.', type=click.Choice(['kernel','deep','gradient']), show_default=True)
@click.option('-ssbs', '--shap_sample_batch_size', default=0, help='Break up shapley computations into batches. Set to 0 for only 1 batch.', show_default=True)
@click.option('-r', '--n_random_representative', default=0, help='Number of representative samples to choose for background selection. Is this number for regression models, or this number per class for classification problems.', show_default=True)
@click.option('-col', '--interest_col', default='disease', help='Column of interest for sample selection for explainer training.', type=click.Path(exists=False), show_default=True)
@click.option('-rt', '--n_random_representative_test', default=0, help='Number of representative samples to choose for test set. Is this number for regression models, or this number per class for classification problems.', show_default=True)
@click.option('-e', '--categorical_encoder', default='./predictions/one_hot_encoder.p', help='One hot encoder if categorical model. If path exists, then return top positive controbutions per samples of that class. Encoded values must be of sample class as interest_col.', type=click.Path(exists=False), show_default=True)
@click.option('-cc', '--cross_class', is_flag=True, help='Find importance of features for prediction across classes.')
@click.option('-fs', '--feature_selection', is_flag=True, help='Perform feature selection using top global SHAP scores.', show_default=True)
@click.option('-top', '--top_outputs', default=0, help='Get shapley values for fewer outputs if feature selection.', show_default=True)
@click.option('-vae', '--vae_interpret', is_flag=True, help='Use model to get shapley values for VAE latent dimensions, only works if proper datasets are set.', show_default=True)
@click.option('-cl', '--pred_class', default='', help='Prediction class top cpgs.', type=click.Path(exists=False), show_default=True)
@click.option('-r', '--results_csv', default='./predictions/results.csv', help='Remove all misclassifications.', type=click.Path(exists=False), show_default=True)
@click.option('-ind', '--individual', default='', help='One individual top cpgs.', type=click.Path(exists=False), show_default=True)
@click.option('-rc', '--residual_cutoff', default=0., multiple=True, help='Activate for regression interpretations. Standard deviation in residuals before removing.', show_default=True)
@click.option('-cn', '--cell_names', default=[''], multiple=True, help='Multioutput names for multi-output regression.', show_default=True)
@click.option('-dsd', '--dump_shap_data', is_flag=True, help='Dump raw SHAP data to numpy file, warning, may be large file.', show_default=True)
def produce_shapley_data(train_pkl, val_pkl, test_pkl, model_pickle, n_workers, batch_size, cuda, n_samples, n_top_features, output_dir, method, shap_sample_batch_size, n_random_representative, interest_col, n_random_representative_test, categorical_encoder, cross_class, feature_selection, top_outputs, vae_interpret, pred_class, results_csv, individual, residual_cutoff, cell_names, dump_shap_data):
    """Explanations (coefficients for CpGs) for every individual prediction.
    Produce SHAPley scores for each CpG for each individualized prediction, and then aggregate them across coarser classes.
    Store CpGs with top SHAPley scores as well for quick access.
    Store in Shapley data object."""
    os.makedirs(output_dir,exist_ok=True)
    if not pred_class:
        pred_class = None
    if not top_outputs or not feature_selection:
        top_outputs = None
    if os.path.exists(categorical_encoder):
        categorical_encoder=pickle.load(open(categorical_encoder,'rb'))
        if categorical_encoder:
            prediction_classes=list(categorical_encoder.categories_[0])
        else:
            prediction_classes = None
    else:
        prediction_classes = None
    cell_names=list(filter(None,cell_names))
    train_methyl_array, val_methyl_array, test_methyl_array=MethylationArray.from_pickle(train_pkl), MethylationArray.from_pickle(val_pkl), MethylationArray.from_pickle(test_pkl)#preprocessed_methyl_array.subset_index(train_test_idx_dict['train']), preprocessed_methyl_array.subset_index(train_test_idx_dict['test'])
    #preprocessed_methyl_array=MethylationArray(*extract_pheno_beta_df_from_pickle_dict(input_dict))
    train_methyl_array = MethylationArrays([train_methyl_array,val_methyl_array]).combine()
    cpgs, train_samples= train_methyl_array.return_cpgs(), train_methyl_array.return_idx()
    cpgs, test_samples= test_methyl_array.return_cpgs(), test_methyl_array.return_idx()
    model = torch.load(model_pickle)
    if os.path.exists(results_csv):
        results_df=pd.read_csv(results_csv)
        if residual_cutoff:
            replace_dict={k:k.replace('_pred','') for k in list(results_df)}
            pred_df = results_df.iloc[:,np.vectorize(lambda x: x.endswith('_pred'))(list(results_df))].rename(replace_dict)
            true_df = results_df.iloc[:,np.vectorize(lambda x: x.endswith('_true'))(list(results_df))].rename(replace_dict)
            residual_df = pred_df-true_df
            cell_names=list(residual_df)
            for col in residual_df:
                residual_df.loc[:,col] = (residual_df[col] - residual_df[col].mean())/residual_df[col].std(ddof=0)
            residual_df = (results_df <= residual_cutoff)
            residual_df = residual_df.all(axis=1)
            test_methyl_array=test_methyl_array.subset_index(np.array(list(residual_df.index)))
        else:
            test_methyl_array=test_methyl_array.subset_index(np.array(list((results_df['y_pred']==results_df['y_true']).index)))
    if individual:
        test_methyl_array=test_methyl_array.subset_index(np.array([individual]))
    if n_random_representative and method != 'gradient':
        train_methyl_array = train_methyl_array.subsample(interest_col, n_samples=n_random_representative, categorical=model.categorical if 'categorical' in dir(model) else False)
    if n_random_representative_test:
        test_methyl_array = test_methyl_array.subsample(interest_col, n_samples=n_random_representative_test, categorical=model.categorical if 'categorical' in dir(model) else False)
    if 'categorical' in dir(model) and model.categorical:
        print("TRAIN:")
        train_methyl_array.categorical_breakdown(interest_col)
        print("TEST:")
        test_methyl_array.categorical_breakdown(interest_col)
    model.eval()
    if cuda:
        model = model.cuda()
    if vae_interpret:
        model.forward = model.mlp
    elif method == 'deep' or method == 'gradient':
            model.forward = model.forward_predict
    t_arr = train_methyl_array.return_raw_beta_array()
    prediction_function=main_prediction_function(n_workers,min(batch_size,t_arr.shape[0]), model, cuda) if method == 'kernel' else model
    # in order for shap to work, prediction_function must work with methylation array beta values
    n_test_results_outputs=main_prediction_function(n_workers, 1, model, cuda)(t_arr[:2, ...]).shape[1]
    # if above does not work, shap will not work
    cpg_explainer = CpGExplainer(prediction_function, cuda)
    cpg_explainer.build_explainer(train_methyl_array, method, batch_size=batch_size)
    if not shap_sample_batch_size:
        shap_sample_batch_size = None
    cpg_explainer.return_shapley_scores(test_methyl_array, n_samples, n_outputs=n_test_results_outputs, shap_sample_batch_size=shap_sample_batch_size, top_outputs=top_outputs)
    if dump_shap_data:
        np.save(file='raw_shap_data.npy',arr=cpg_explainer.shap_values)
    if not residual_cutoff:
        cpg_explainer.classifier_assign_scores_to_shap_data(test_methyl_array, n_top_features, interest_col=interest_col, prediction_classes=prediction_classes)
    else:
        cpg_explainer.regressor_assign_scores_to_shap_data(test_methyl_array, n_top_features, cell_names=cell_names)
    if feature_selection:
        print('FEATURE SELECT')
        feature_selected_methyl_array=cpg_explainer.feature_select(MethylationArrays([train_methyl_array,test_methyl_array]).combine(),n_top_features)
        feature_selected_methyl_array.write_pickle(join(output_dir, 'feature_selected_methyl_array.pkl'))
    else:
        output_all_cpgs=join(output_dir,'all_cpgs.p')
        shapley_output=join(output_dir,'shapley_data.p')
        pickle.dump(train_methyl_array.return_cpgs(),open(output_all_cpgs,'wb'))
        cpg_explainer.shapley_data.to_pickle(shapley_output)

@interpret.command()
@click.option('-o', '--output_dir', default='./lola_db/', help='Output directory for lola dbs.', type=click.Path(exists=False), show_default=True)
def grab_lola_db_cache(output_dir):
    """Download core and extended LOLA databases for enrichment tests."""
    os.makedirs(output_dir,exist_ok=True)
    core_dir = join(output_dir,'core')
    extended_dir = join(output_dir,'extended')
    subprocess.call("wget http://big.databio.org/regiondb/LOLACoreCaches_180412.tgz && wget http://big.databio.org/regiondb/LOLAExtCaches_170206.tgz && mv *.tgz {0} && tar -xvzf {0}/LOLAExtCaches_170206.tgz && mv {0}/scratch {1} && tar -xvzf {0}/LOLACoreCaches_180412.tgz && mv {0}/nm {2}".format(output_dir,extended_dir,core_dir),shell=True)

@interpret.command()
@click.option('-a', '--all_cpgs_pickle', default='./interpretations/shapley_explanations/all_cpgs.p', help='List of all cpgs used in shapley analysis.', type=click.Path(exists=False), show_default=True)
@click.option('-s', '--shapley_data_list', default=['./interpretations/shapley_explanations/shapley_data.p'], multiple=True, help='Pickle containing top CpGs.', type=click.Path(exists=False), show_default=True)
@click.option('-o', '--output_dir', default='./interpretations/biological_explanations/', help='Output directory for interpretations.', type=click.Path(exists=False), show_default=True)
@click.option('-w', '--analysis', default='GO', help='Choose biological analysis.', type=click.Choice(['GSEA','GO','KEGG','GENE', 'LOLA', 'NEAR_CPGS']), show_default=True)
@click.option('-n', '--n_workers', default=8, help='Number workers.', show_default=True)
@click.option('-l', '--lola_db', default='./lola_db/core/nm/t1/resources/regions/LOLACore/hg19/', help='LOLA region db.', type=click.Path(exists=False), show_default=True)
@click.option('-i', '--individuals', default=[''], multiple=True, help='Individuals to evaluate.', show_default=True)
@click.option('-c', '--classes', default=[''], multiple=True, help='Classes to evaluate.', show_default=True)
@click.option('-m', '--max_gap', default=1000, help='Genomic distance to search for nearby CpGs than found top cpgs shapleys.', show_default=True)
@click.option('-ci', '--class_intersect', is_flag=True, help='Compute class shapleys by intersection of individuals.', show_default=True)
@click.option('-lo', '--length_output', default=20, help='Number enriched terms to print.', show_default=True)
@click.option('-ss', '--set_subtraction', is_flag=True, help='Only consider CpGs relevant to particular class.', show_default=True)
@click.option('-int', '--intersection', is_flag=True, help='CpGs common to all classes.', show_default=True)
@click.option('-col', '--collections', default=[''], multiple=True, help='Lola collections.', type=click.Choice(['','cistrome_cistrome','codex','encode_tfbs,ucsc_features','cistrome_epigenome','encode_segmentation','sheffield_dnase','jaspar_motifs','roadmap_epigenomics']), show_default=True)
@click.option('-gsea', '--gsea_analyses', default=[''], multiple=True, help='Gene set enrichment analysis to choose from, if chosen will override other analysis options. http://software.broadinstitute.org/gsea/msigdb/collections.jsp', type=click.Choice(['','C1','C2','C2.KEGG,C2.REACTOME','C3','C3.TFT','C4.CGN','C4.CM','C4','C5','C6','C7','C8','H']), show_default=True)
@click.option('-gsp', '--gsea_pickle', default='./data/gsea_collections.p', help='Gene set enrichment analysis to choose from.', type=click.Path(exists=False), show_default=True)
@click.option('-d', '--depletion', is_flag=True, help='Run depletion LOLA analysis instead of enrichment.', show_default=True)
@click.option('-ov', '--overlap_test', is_flag=True, help='Run overlap test with IDOL library instead of all other tests.', show_default=True)
@click.option('-cgs', '--cpg_set', default='IDOL', help='Test for clock or IDOL enrichments.', type=click.Choice(['IDOL','epitoc','horvath','hannum']), show_default=True)
@click.option('-g', '--top_global', is_flag=True, help='Look at global top cpgs, overwrites classes and individuals.', show_default=True)
@click.option('-ex', '--extract_library', is_flag=True, help='Extract a library of cpgs to subset future array for inspection or prediction tests.', show_default=True)
def interpret_biology(all_cpgs_pickle,shapley_data_list,output_dir, analysis, n_workers, lola_db, individuals, classes, max_gap, class_intersect, length_output, set_subtraction, intersection, collections, gsea_analyses, gsea_pickle, depletion, overlap_test, cpg_set, top_global, extract_library):
    """Interrogate CpGs with high SHAPley scores for individuals, classes, or overall, for enrichments, genes, GO, KEGG, LOLA, overlap with popular cpg sets, GSEA, find nearby cpgs to top."""
    os.makedirs(output_dir,exist_ok=True)
    if os.path.exists(all_cpgs_pickle):
        all_cpgs=list(pickle.load(open(all_cpgs_pickle,'rb')))
    else:
        all_cpgs = []
    #top_cpgs=pickle.load(open(top_cpgs_pickle,'rb'))
    head_individuals=list(filter(None,individuals))
    head_classes=list(filter(None,classes))
    collections=list(filter(None,collections))
    gsea_analyses=list(filter(None,gsea_analyses))
    if gsea_analyses:
        analysis='GSEA'
    for i,shapley_data in enumerate(shapley_data_list):
        shapley_data=ShapleyData.from_pickle(shapley_data)
        shapley_data_explorer=ShapleyDataExplorer(shapley_data)
        if not top_global:
            if head_classes and head_classes[0]=='all':
                classes = shapley_data_explorer.list_classes()
            else:
                classes = copy.deepcopy(head_classes)
            if head_individuals and head_individuals[0]=='all':
                individuals = shapley_data_explorer.list_individuals(return_list=True)
            else:
                individuals = copy.deepcopy(head_individuals)
            cpg_exclusion_sets,cpg_sets=None,None
            if set_subtraction:
                _,cpg_exclusion_sets=shapley_data_explorer.return_cpg_sets()
            elif intersection:
                cpg_sets,_=shapley_data_explorer.return_cpg_sets()
            top_cpgs=shapley_data_explorer.return_top_cpgs(classes=classes,individuals=individuals,cpg_exclusion_sets=cpg_exclusion_sets,cpg_sets=cpg_sets)
        else:
            top_cpgs = {'global':shapley_data_explorer.return_global_importance_cpgs()}
        bio_interpreter = BioInterpreter(top_cpgs)
        if overlap_test:
            output_library_cpgs=bio_interpreter.return_overlap_score(set_cpgs=cpg_set,all_cpgs=all_cpgs,output_csv=join(output_dir,'{}_overlaps.csv'.format(cpg_set)),extract_library=extract_library)
            if output_library_cpgs is not None:
                pickle.dump(output_library_cpgs,open(join(output_dir,'cpg_library.pkl'),'wb'))
        else:
            if analysis in ['GO','KEGG','GENE']:
                analysis_outputs=bio_interpreter.gometh(analysis, allcpgs=all_cpgs, length_output=length_output)
            elif analysis == 'LOLA':
                analysis_outputs=bio_interpreter.run_lola(all_cpgs=all_cpgs, lola_db=lola_db, cores=n_workers, collections=collections, depletion=depletion)
            elif analysis == 'NEAR_CPGS':
                analysis_outputs=bio_interpreter.get_nearby_cpg_shapleys(all_cpgs=all_cpgs,max_gap=max_gap, class_intersect=class_intersect)
            elif gsea_analyses:
                analysis_outputs=bio_interpreter.gometh('', allcpgs=all_cpgs, length_output=length_output, gsea_analyses=gsea_analyses, gsea_pickle=gsea_pickle)
            for k in analysis_outputs:
                output_csv=join(output_dir,'{}_{}_{}.csv'.format(shapley_data_list[i].split('/')[-1],k,analysis if not gsea_analyses else '_'.join(gsea_analyses)).replace('/','-'))
                analysis_outputs[k].to_csv(output_csv)

@interpret.command()
@click.option('-s', '--shapley_data', default='./interpretations/shapley_explanations/shapley_data.p', help='Pickle containing top CpGs.', type=click.Path(exists=False), show_default=True)
@click.option('-co', '--classes_only', is_flag=True, help='Only take top CpGs from each class.', show_default=True)
@click.option('-o', '--output_dir', default='./interpretations/shapley_explanations/top_cpgs_extracted_methylarr/', help='Output directory for methylation array.', type=click.Path(exists=False), show_default=True)
@click.option('-t', '--test_pkl', default='./train_val_test_sets/test_methyl_array.pkl', help='Pickle containing testing set.', type=click.Path(exists=False), show_default=True)
@click.option('-c', '--col', default='', help='Column to color for output csvs.', show_default=True)
@click.option('-g', '--global_vals', is_flag=True, help='Only take top CpGs globally.', show_default=True)
@click.option('-n', '--n_extract', default=1000, help='Number cpgs to extract.', show_default=True)
def extract_methylation_array(shapley_data, classes_only, output_dir, test_pkl, col, global_vals, n_extract):
    """Subset and write methylation array using top SHAP cpgs."""
    os.makedirs(output_dir,exist_ok=True)
    shapley_data=ShapleyData.from_pickle(shapley_data)
    shapley_data_explorer=ShapleyDataExplorer(shapley_data)
    output_methyl_arr = shapley_data_explorer.extract_methylation_array(MethylationArray.from_pickle(test_pkl),classes_only,global_vals=global_vals,n_extract=n_extract)
    output_methyl_arr.write_pickle(os.path.join(output_dir,'extracted_methyl_arr.pkl'))
    if col:
        output_methyl_arr.beta[col] = output_methyl_arr.pheno[col]
    output_methyl_arr.write_csvs(output_dir)

@interpret.command()
@click.option('-l', '--head_lola_dir', default='interpretations/biological_explanations/', help='Location of lola output csvs.', type=click.Path(exists=False), show_default=True)
@click.option('-o', '--plot_output_dir', default='./interpretations/biological_explanations/lola_plots/', help='Output directory for interpretations.', type=click.Path(exists=False), show_default=True)
@click.option('-d', '--description_col', default='description', help='Description column for categorical variables', type=click.Path(exists=False),show_default=True)
@click.option('-c', '--cell_types', default=[''], multiple=True, help='Cell types.', show_default=True)
@click.option('-ov', '--overwrite', is_flag=True, help='Overwrite existing plots.', show_default=True)
def plot_all_lola_outputs(head_lola_dir,plot_output_dir,description_col,cell_types,overwrite):
    """Iterate through all LOLA csv results and plot forest plots of all enrichment scores."""
    import glob
    for lola_csv in glob.iglob(join(head_lola_dir,'**','*_LOLA.csv'),recursive=True):
        output_dir=join(plot_output_dir,lola_csv.split('/')[-2])
        if overwrite and len(glob.glob(join(output_dir,'{}*.png'.format(lola_csv.split('/')[-1].replace('.csv',''))))):
            continue
        else:
            plot_lola_output_(lola_csv, output_dir, description_col, cell_types)

@interpret.command()
@click.option('-l', '--lola_csv', default='', help='Location of lola output csv.', type=click.Path(exists=False), show_default=True)
@click.option('-o', '--plot_output_dir', default='./interpretations/biological_explanations/lola_plots/', help='Output directory for interpretations.', type=click.Path(exists=False), show_default=True)
@click.option('-d', '--description_col', default='description', help='Description column for categorical variables', type=click.Path(exists=False),show_default=True)
@click.option('-c', '--cell_types', default=[''], multiple=True, help='Cell types.', show_default=True)
def plot_lola_output(lola_csv, plot_output_dir, description_col, cell_types):
    """Plot LOLA results via forest plot for one csv file."""
    plot_lola_output_(lola_csv, plot_output_dir, description_col, cell_types)

@interpret.command()
@click.option('-s', '--shapley_data', default='./interpretations/shapley_explanations/shapley_data.p', help='Pickle containing top CpGs.', type=click.Path(exists=False), show_default=True)
@click.option('-o', '--output_dir', default='./interpretations/shapley_explanations/shapley_data_by_methylation/', help='Output directory for hypo/hyper shap data.', type=click.Path(exists=False), show_default=True)
@click.option('-t', '--test_pkl', default='./train_val_test_sets/test_methyl_array.pkl', help='Pickle containing testing set.', type=click.Path(exists=False), show_default=True)
@click.option('-thr', '--threshold', default='original', help='Pickle containing testing set.', type=click.Choice(['original','mean']), show_default=True)
def split_hyper_hypo_methylation(shapley_data, output_dir, test_pkl, threshold):
    """Split SHAPleyData object by methylation type (needs to change; but low methylation is 0.5 beta value and below) and output to file."""
    os.makedirs(output_dir,exist_ok=True)
    shapley_data=ShapleyData.from_pickle(shapley_data)
    shapley_data_explorer=ShapleyDataExplorer(shapley_data)
    test_methyl_array=MethylationArray.from_pickle(test_pkl)
    shap_data_methylation=shapley_data_explorer.return_shapley_data_by_methylation_status(test_methyl_array, threshold)
    shap_data_methylation['hypo'].to_pickle(join(output_dir,'hypo_shapley_data.p'))
    shap_data_methylation['hyper'].to_pickle(join(output_dir,'hyper_shapley_data.p'))

@interpret.command()
@click.option('-s', '--shapley_data', default='./interpretations/shapley_explanations/shapley_data.p', help='Pickle containing top CpGs.', type=click.Path(exists=False), show_default=True)
@click.option('-t', '--test_pkl', default='./train_val_test_sets/test_methyl_array.pkl', help='Pickle containing testing set.', type=click.Path(exists=False), show_default=True)
@click.option('-c', '--col', default='age', help='Column to turn into bins.', type=click.Path(exists=False),show_default=True)
@click.option('-n', '--n_bins', default=10, help='Number of bins.',show_default=True)
@click.option('-ot', '--output_test_pkl', default='./train_val_test_sets/test_methyl_array_shap_binned.pkl', help='Binned shap pickle for further testing.', type=click.Path(exists=False), show_default=True)
@click.option('-os', '--output_shap_pkl', default='./interpretations/shapley_explanations/shapley_binned.p', help='Pickle containing top CpGs, binned phenotype.', type=click.Path(exists=False), show_default=True)
def bin_regression_shaps(shapley_data, test_pkl,col,n_bins,output_test_pkl,output_shap_pkl):
    """Take aggregate individual scores from regression results, that normally are not categorized, and aggregate across new category created from continuous data."""
    os.makedirs(output_test_pkl[:output_test_pkl.rfind('/')],exist_ok=True)
    os.makedirs(output_shap_pkl[:output_shap_pkl.rfind('/')],exist_ok=True)
    shapley_data=ShapleyData.from_pickle(shapley_data)
    shapley_data_explorer=ShapleyDataExplorer(shapley_data)
    test_methyl_array=MethylationArray.from_pickle(test_pkl)
    new_col_name = test_methyl_array.bin_column(col,n_bins)
    print(test_methyl_array.categorical_breakdown(new_col_name))
    shapley_data = shapley_data_explorer.return_binned_shapley_data(col, test_methyl_array.pheno[new_col_name], add_top_negative=False)
    print(ShapleyDataExplorer(shapley_data).list_classes())
    test_methyl_array.write_pickle(output_test_pkl)
    shapley_data.to_pickle(output_shap_pkl)

# create force plots for each sample??? use output shapley values and output explainer
@interpret.command()
@click.option('-s', '--shapley_data', default='./interpretations/shapley_explanations/shapley_data.p', help='Pickle containing top CpGs.', type=click.Path(exists=False), show_default=True)
@click.option('-c', '--class_names', default=[''], multiple=True, help='Class names.', show_default=True)
@click.option('-o', '--output_dir', default='./interpretations/shapley_explanations/top_cpgs_jaccard/', help='Output directory for cpg jaccard_stats.', type=click.Path(exists=False), show_default=True)
@click.option('-ov', '--overall', is_flag=True, help='Output overall similarity.', show_default=True)
@click.option('-i', '--include_individuals', is_flag=True, help='Output individuals.', show_default=True)
@click.option('-co', '--cooccurrence', is_flag=True, help='Output cooccurrence instead jaccard.', show_default=True)
@click.option('-opt', '--optimize_n_cpgs', is_flag=True, help='Search for number of top CpGs to use.', show_default=True)
def shapley_jaccard(shapley_data,class_names, output_dir, overall, include_individuals, cooccurrence, optimize_n_cpgs):
    """Plot Shapley Jaccard Similarity Matrix to demonstrate sharing of Top CpGs between classes/groupings of individuals."""
    os.makedirs(output_dir,exist_ok=True)
    shapley_data=ShapleyData.from_pickle(shapley_data)
    shapley_data_explorer=ShapleyDataExplorer(shapley_data)
    outfilename=join(output_dir,'{}_jaccard.csv'.format('_'.join(class_names)))
    shapley_data_explorer.jaccard_similarity_top_cpgs(class_names,include_individuals,overall, cooccurrence).to_csv(outfilename)
    if optimize_n_cpgs:
        from skopt import gp_minimize
        shapley_data_explorer2 = ShapleyDataExplorer(shapley_data)
        def f(x):
            shapley_data_explorer2.shapley_data = shapley_data_explorer.limit_number_top_cpgs(n_top_cpgs=x[0])
            jaccard_matrix = shapley_data_explorer2.jaccard_similarity_top_cpgs(class_names,include_individuals,overall, cooccurrence).values
            jaccard_mean=jaccard_matrix.mean()
            print(x[0],jaccard_mean)
            return -jaccard_mean
        results=gp_minimize(f,[(100,shapley_data_explorer.shapley_data.top_cpgs['overall'].shape[0])],n_calls=25,n_jobs=8)
        print("{} number CpGs produces most similar jaccard similarity matrix.".format(results.x))
        print(dict(zip(results.x_iters.flatten().tolist(), (results.func_vals*-1).tolist())))

@interpret.command()
@click.option('-i', '--input_csv', default='./interpretations/shapley_explanations/top_cpgs_jaccard/all_jaccard.csv', help='Output directory for cpg jaccard_stats.', type=click.Path(exists=False), show_default=True)
@click.option('-t', '--test_pkl', default='./train_val_test_sets/test_methyl_array.pkl', help='Pickle containing testing set.', type=click.Path(exists=False), show_default=True)
@click.option('-c', '--col', default='disease', help='Column to sort on.', show_default=True)
@click.option('-o', '--output_csv', default='./interpretations/shapley_explanations/top_cpgs_jaccard/all_jaccard_sorted.csv', help='Output directory for cpg jaccard_stats.', type=click.Path(exists=False), show_default=True)
@click.option('-sym', '--symmetric', is_flag=True, help='Is symmetric?', show_default=True)
def order_results_by_col(input_csv, test_pkl, col, output_csv, symmetric):
    """Order results that produces some CSV by phenotype column, alphabetical ordering to show maybe that results group together. Plot using pymethyl-visualize."""
    os.makedirs(output_csv[:output_csv.rfind('.')],exist_ok=True)
    df=pd.read_csv(input_csv,index_col=0)
    if os.path.exists(test_pkl):
        test_arr=MethylationArray.from_pickle(test_pkl) # temporary solution
        test_arr_idx=test_arr.pheno.sort_values(col).index.values
        if np.all(np.vectorize(lambda x: x.startswith(col))(df.index.values)):
            test_arr_idx = np.vectorize(lambda x: '{}_{}'.format(col,x))(test_arr_idx)
        test_arr_idx=test_arr_idx[np.isin(test_arr_idx,df.index.values)] # temporary fix
    else:
        test_arr_idx = np.sort(df.index.values)
    df=df.loc[test_arr_idx,test_arr_idx] if symmetric else df[test_arr_idx]
    df.to_csv(output_csv)


@interpret.command()
@click.option('-s', '--shapley_data', default='./interpretations/shapley_explanations/shapley_data.p', help='Pickle containing top CpGs.', type=click.Path(exists=False), show_default=True)
@click.option('-i', '--individuals', default=[''], multiple=True, help='Individuals.', show_default=True)
@click.option('-o', '--output_dir', default='./interpretations/shapley_explanations/top_cpgs_methylation/', help='Output directory for cpg methylation shap.', type=click.Path(exists=False), show_default=True)
@click.option('-t', '--test_pkl', default='./train_val_test_sets/test_methyl_array.pkl', help='Pickle containing testing set.', type=click.Path(exists=False), show_default=True)
def view_methylation_top_cpgs(shapley_data,individuals, output_dir, test_pkl):
    """Write the Top CpGs for each class/individuals to file along with methylation information."""
    os.makedirs(output_dir,exist_ok=True)
    shapley_data=ShapleyData.from_pickle(shapley_data)
    shapley_data_explorer=ShapleyDataExplorer(shapley_data)
    test_methyl_array=MethylationArray.from_pickle(test_pkl)
    for individual in individuals:
        class_name,individual,shap_df=shapley_data_explorer.view_methylation(individual=individual,methyl_arr=test_methyl_array)
        shap_df.to_csv(join(output_dir,"{}_{}.csv".format(class_name,individual)))

@interpret.command()
@click.option('-s', '--shapley_data', default='./interpretations/shapley_explanations/shapley_data.p', help='Pickle containing top CpGs.', type=click.Path(exists=False), show_default=True)
def list_individuals(shapley_data):
    """List individuals that have ShapleyData. Not all made the cut from the test dataset."""
    shapley_data=ShapleyData.from_pickle(shapley_data)
    shapley_data_explorer=ShapleyDataExplorer(shapley_data)
    print(shapley_data_explorer.list_individuals())

@interpret.command()
@click.option('-s', '--shapley_data', default='./interpretations/shapley_explanations/shapley_data.p', help='Pickle containing top CpGs.', type=click.Path(exists=False), show_default=True)
def list_classes(shapley_data):
    """List classes/multioutput regression cell-types that have SHAPley data to be interrogated."""
    shapley_data=ShapleyData.from_pickle(shapley_data)
    print(list(shapley_data.top_cpgs['by_class'].keys()))

@interpret.command()
@click.option('-s', '--shapley_data', default='./interpretations/shapley_explanations/shapley_data.p', help='Pickle containing top CpGs.', type=click.Path(exists=False), show_default=True)
@click.option('-nf', '--n_top_features', default=500, show_default=True, help='Top features to select for shap outputs.')
@click.option('-o', '--output_pkl', default='./interpretations/shapley_explanations/shapley_reduced_data.p', help='Pickle containing top CpGs, reduced number.', type=click.Path(exists=False), show_default=True)
def reduce_top_cpgs(shapley_data,n_top_features,output_pkl):
    """Reduce set of top cpgs."""
    os.makedirs(output_pkl[:output_pkl.rfind('/')],exist_ok=True)
    shapley_data=ShapleyData.from_pickle(shapley_data)
    shapley_data_explorer=ShapleyDataExplorer(shapley_data)
    shapley_data_explorer.limit_number_top_cpgs(n_top_cpgs=n_top_features).to_pickle(output_pkl)

@interpret.command()
@click.option('-s', '--shapley_data', default='./interpretations/shapley_explanations/shapley_data.p', help='Pickle containing top CpGs.', type=click.Path(exists=False), show_default=True)
@click.option('-nf', '--n_top_features', default=500, show_default=True, help='Top features to select for shap outputs.')
@click.option('-o', '--output_pkl', default='./interpretations/shapley_explanations/shapley_reduced_data.p', help='Pickle containing top CpGs, reduced number.', type=click.Path(exists=False), show_default=True)
@click.option('-a', '--abs_val', is_flag=True, help='Top CpGs found using absolute value.')
@click.option('-n', '--neg_val', is_flag=True, help='Return top CpGs that are making negative contributions.')
def regenerate_top_cpgs(shapley_data,n_top_features,output_pkl, abs_val, neg_val):
    """Increase size of Top CpGs using the original SHAP scores."""
    os.makedirs(output_pkl[:output_pkl.rfind('/')],exist_ok=True)
    shapley_data=ShapleyData.from_pickle(shapley_data)
    shapley_data_explorer=ShapleyDataExplorer(shapley_data)
    shapley_data_explorer.regenerate_individual_shap_values(n_top_cpgs=n_top_features,abs_val=abs_val, neg_val=neg_val).to_pickle(output_pkl)

@interpret.command()
@click.option('-s', '--shapley_data', default='./interpretations/shapley_explanations/shapley_data.p', help='Pickle containing top CpGs.', type=click.Path(exists=False), show_default=True)
@click.option('-o', '--output_dir', default='./interpretations/shap_outputs/', help='Output directory for output plots.', type=click.Path(exists=False), show_default=True)
@click.option('-i', '--individuals', default=[''], multiple=True, help='Individuals to evaluate.', show_default=True)
@click.option('-c', '--classes', default=[''], multiple=True, help='Classes to evaluate.', show_default=True)
@click.option('-hist', '--output_histogram', is_flag=True, help='Whether to output a histogram for each class/individual of their SHAP scores.')
@click.option('-abs', '--absolute', is_flag=True, help='Use sums of absolute values in making computations.')
@click.option('-log', '--log_transform', is_flag=True, help='Log transform final results for plotting.')
def return_shap_values(shapley_data,output_dir,individuals,classes, output_histogram, absolute, log_transform):
    """Return matrix of shapley values per class, with option to, classes/individuals are columns, CpGs are rows, option to plot multiple histograms/density plots."""
    from sklearn.metrics import pairwise_distances
    os.makedirs(output_dir,exist_ok=True)
    shapley_data=ShapleyData.from_pickle(shapley_data)
    shapley_data_explorer=ShapleyDataExplorer(shapley_data)
    individuals=list(filter(None,individuals))
    classes=list(filter(None,classes))
    if absolute:
        shapley_data_explorer.make_shap_scores_abs()
    if classes and classes[0]=='all':
        classes = shapley_data_explorer.list_classes()
    if individuals and individuals[0]=='all':
        individuals = shapley_data_explorer.list_individuals(return_list=True)
    concat_list=[]
    if classes:
        for class_name in classes:
            concat_list.append(shapley_data_explorer.extract_class(class_name,get_shap_values=True))
    if individuals:
        for individual in individuals:
            concat_list.append(shapley_data_explorer.extract_individual(individual,get_shap_values=True))
    df=pd.concat(concat_list,axis=1,keys=classes+individuals)
    df.to_csv(join(output_dir,'returned_shap_values.csv'))
    pd.DataFrame(pairwise_distances(df.T.values,metric='correlation'),index=list(df),columns=list(df)).to_csv(join(output_dir,'returned_shap_values_corr_dist.csv'))
    #df.corr().to_csv(join(output_dir,'returned_shap_values_corr_dist.csv'))
    if output_histogram:
        import matplotlib
        matplotlib.use('Agg')
        import matplotlib.pyplot as plt
        import seaborn as sns
        sns.set()
        for entity in (classes+individuals):
            shap_scores = df[entity]
            plt.figure()
            ax=sns.distplot(shap_scores, kde=(log_transform==False))
            if log_transform:
                ax.set_xscale('symlog')
            plt.xlabel('{} SHAP value'.format("Log" if log_transform else ""))
            plt.ylabel('Frequency')
            plt.savefig(join(output_dir,'{}_shap_values.png'.format(entity)),dpi=300)
            top_10_shaps = shap_scores.abs().sort_values(ascending=False).iloc[:10]
            top_10_shaps=pd.DataFrame({'cpgs':top_10_shaps.index, '|SHAP|':top_10_shaps.values})
            plt.figure()
            ax=sns.barplot('|SHAP|','cpgs',orient='h',data=top_10_shaps)
            ax.tick_params(labelsize=4)
            plt.savefig(join(output_dir,'{}_top_shap_values.png'.format(entity)),dpi=300)
    if not absolute:
        print("All saved values reflect absolute values of sums, not sums of absolute values, if CpG SHAP scores are opposite signs across individuals, this will reduce the score of the resulting SHAP estimate.")

@interpret.command()
@click.option('-s', '--shapley_data', default='./interpretations/shapley_explanations/shapley_data.p', help='Pickle containing top CpGs.', type=click.Path(exists=False), show_default=True)
@click.option('-o', '--output_dir', default='./interpretations/output_plots/', help='Output directory for output plots.', type=click.Path(exists=False), show_default=True)
@click.option('-i', '--individuals', default=[''], multiple=True, help='Individuals to evaluate.', show_default=True)
@click.option('-c', '--classes', default=[''], multiple=True, help='Classes to evaluate.', show_default=True)
def plot_top_cpgs(shapley_data,output_dir,individuals,classes):
    """Plot the top cpgs locations in the genome using circos."""
    os.makedirs(output_dir,exist_ok=True)
    shapley_data=ShapleyData.from_pickle(shapley_data)
    shapley_data_explorer=ShapleyDataExplorer(shapley_data)
    individuals=list(filter(None,individuals))
    classes=list(filter(None,classes))
    if classes and classes[0]=='all':
        classes = shapley_data_explorer.list_classes()
    if individuals and individuals[0]=='all':
        individuals = shapley_data_explorer.list_individuals(return_list=True)
    cpg_exclusion_sets,cpg_sets=None,None
    set_subtraction,intersection=False,False
    if set_subtraction:
        _,cpg_exclusion_sets=shapley_data_explorer.return_cpg_sets()
    elif intersection:
        cpg_sets,_=shapley_data_explorer.return_cpg_sets()
    top_cpgs=shapley_data_explorer.return_top_cpgs(classes=classes,individuals=individuals,cpg_exclusion_sets=cpg_exclusion_sets,cpg_sets=cpg_sets)
    circos_plotter=PlotCircos()
    circos_plotter.plot_cpgs(top_cpgs,output_dir)

@interpret.command()
@click.option('-i', '--embedding_methyl_array_pkl', default='./embeddings/vae_methyl_arr.pkl', help='Use ./predictions/vae_mlp_methyl_arr.pkl or ./embeddings/vae_methyl_arr.pkl for vae interpretations.', type=click.Path(exists=False), show_default=True)
@click.option('-c', '--pheno_col', default='disease', help='Column to separate on.', type=click.Path(exists=False), show_default=True)
@click.option('-m', '--metric', default='cosine', help='Distance metric to compare classes.', type=click.Path(exists=False), show_default=True)
@click.option('-t', '--trim', default=0.0, help='Trim outlier distances. Number 0-0.5.', show_default=True)
@click.option('-o', '--output_csv', default='./results/class_embedding_differences.csv', help='Distances between classes.', type=click.Path(exists=False), show_default=True)
@click.option('-op', '--output_pval_csv', default='', help='If specify a CSV file, outputs pairwise manova tests between embeddings for clusters.', type=click.Path(exists=False), show_default=True)
def interpret_embedding_classes(embedding_methyl_array_pkl, pheno_col, metric, trim, output_csv, output_pval_csv):
    """Compare average distance between classes in embedding space."""
    os.makedirs(output_csv[:output_csv.rfind('/')],exist_ok=True)
    methyl_array = MethylationArray.from_pickle(embedding_methyl_array_pkl)
    distance_compute = DistanceMatrixCompute(methyl_array, pheno_col)
    distance_compute.compute_distances(metric, trim)
    distance_compute.return_distances().to_csv(output_csv)
    if output_pval_csv:
        distance_compute.calculate_p_values()
        distance_compute.return_p_values().to_csv(output_pval_csv)

#################

if __name__ == '__main__':
    interpret()



import click
import os, copy
from os.path import join
import subprocess

CONTEXT_SETTINGS = dict(help_option_names=['-h','--help'], max_content_width=90)

@click.group(context_settings= CONTEXT_SETTINGS)
@click.version_option(version='0.1')
def test():
    pass

@test.command()
def test_pipeline():
    print("WARNING: Make sure to download test data from https://github.com/Christensen-Lab-Dartmouth/MethylNet, unzip files/folder and make sure they're located in train_val_test_sets.")
    commands = """mkdir visualizations results
methylnet-embed launch_hyperparameter_scan --hyperparameter_yaml example_embedding_hyperparameter_grid.yaml -sc Age -mc 0.84 -b 1. -g -j 20
methylnet-embed launch_hyperparameter_scan --hyperparameter_yaml example_embedding_hyperparameter_grid.yaml -sc Age -g -n 1 -b 1.
pymethyl-visualize transform_plot -i embeddings/vae_methyl_arr.pkl -nn 8 -c Age -o results/vae_embedding_plot.html
methylnet-predict launch_hyperparameter_scan --hyperparameter_yaml example_prediction_hyperparameter_grid.yaml -ic Age -mc 0.84 -g -j 20
methylnet-predict launch_hyperparameter_scan --hyperparameter_yaml example_prediction_hyperparameter_grid.yaml -ic Age -g -n 1
pymethyl-visualize transform_plot -i predictions/vae_mlp_methyl_arr.pkl -nn 8 -c Age -o results/mlp_embedding_plot.html
methylnet-predict regression_report
methylnet-visualize plot_training_curve -t embeddings/training_val_curve.p -vae -o results/embed_training_curve.png -thr 2e8
methylnet-visualize plot_training_curve -thr 2e6""".splitlines()
    for command in commands:
        subprocess.call(command,shell=True)
    print("Check results in embeddings, predictions, visualizations, results.")

#################

if __name__ == '__main__':
    test()



"""
schedulers.py
=======================
Learning rate schedulers that help enable better and more generalizable models."""

import torch
from torch.optim.lr_scheduler import ExponentialLR,LambdaLR
import math


class CosineAnnealingWithRestartsLR(torch.optim.lr_scheduler._LRScheduler):
    r"""Borrowed from: https://github.com/mpyrozhok/adamwr/blob/master/cyclic_scheduler.py
    Needs to be updated to reflect newest changes.
    From original docstring:
    Set the learning rate of each parameter group using a cosine annealing
    schedule, where :math:`\eta_{max}` is set to the initial lr and
    :math:`T_{cur}` is the number of epochs since the last restart in SGDR:
     .. math::
         \eta_t = \eta_{min} + \frac{1}{2}(\eta_{max} - \eta_{min})(1 +
        \cos(\frac{T_{cur}}{T_{max}}\pi))
     When last_epoch=-1, sets initial lr as lr.
     It has been proposed in
    `SGDR: Stochastic Gradient Descent with Warm Restarts`_. This implements
    the cosine annealing part of SGDR, the restarts and number of iterations multiplier.
     Args:
        optimizer (Optimizer): Wrapped optimizer.
        T_max (int): Maximum number of iterations.
        T_mult (float): Multiply T_max by this number after each restart. Default: 1.
        eta_min (float): Minimum learning rate. Default: 0.
        last_epoch (int): The index of last epoch. Default: -1.
     .. _SGDR\: Stochastic Gradient Descent with Warm Restarts:
        https://arxiv.org/abs/1608.03983
    """
    def __init__(self, optimizer, T_max, eta_min=0, last_epoch=-1, T_mult=1., alpha_decay=1.0):
        self.T_max = T_max
        self.T_mult = T_mult
        self.restart_every = T_max
        self.eta_min = eta_min
        self.restarts = 0
        self.restarted_at = 0
        self.alpha = alpha_decay
        super().__init__(optimizer, last_epoch)

    def restart(self):
        self.restarts += 1
        self.restart_every = int(round(self.restart_every * self.T_mult))
        self.restarted_at = self.last_epoch

    def cosine(self, base_lr):
        return self.eta_min + self.alpha**self.restarts * (base_lr - self.eta_min) * (1 + math.cos(math.pi * self.step_n / self.restart_every)) / 2

    @property
    def step_n(self):
        return self.last_epoch - self.restarted_at

    def get_lr(self):
        if self.step_n >= self.restart_every:
            self.restart()
        return [self.cosine(base_lr) for base_lr in self.base_lrs]

class Scheduler:
    """Scheduler class that modulates learning rate of torch optimizers over epochs.

    Parameters
    ----------
    optimizer : type
        torch.Optimizer object
    opts : type
        Options of setting the learning rate scheduler, see default.

    Attributes
    ----------
    schedulers : type
        Different types of schedulers to choose from.
    scheduler_step_fn : type
        How scheduler updates learning rate.
    initial_lr : type
        Initial set learning rate.
    scheduler_choice : type
        What scheduler type was chosen.
    scheduler : type
        Scheduler object chosen that will more directly update optimizer LR.

    """

    def __init__(self, optimizer=None, opts=dict(scheduler='null',lr_scheduler_decay=0.5,T_max=10,eta_min=5e-8,T_mult=2)):
        self.schedulers = {'exp':(lambda optimizer: ExponentialLR(optimizer, opts["lr_scheduler_decay"])),
                            'null':(lambda optimizer: None),
                            'warm_restarts':(lambda optimizer: CosineAnnealingWithRestartsLR(optimizer, T_max=opts['T_max'], eta_min=opts['eta_min'], last_epoch=-1, T_mult=opts['T_mult']))}
        self.scheduler_step_fn = {'exp':(lambda scheduler: scheduler.step()),
                                  'warm_restarts':(lambda scheduler: scheduler.step()),
                                  'null':(lambda scheduler: None)}
        self.initial_lr = optimizer.param_groups[0]['lr']
        self.scheduler_choice = opts['scheduler']
        self.scheduler = self.schedulers[self.scheduler_choice](optimizer) if optimizer is not None else None

    def step(self):
        """Update optimizer learning rate"""
        self.scheduler_step_fn[self.scheduler_choice](self.scheduler)

    def get_lr(self):
        """Return current learning rate.

        Returns
        -------
        float
            Current learning rate.

        """
        lr = (self.initial_lr if self.scheduler_choice == 'null' else self.scheduler.optimizer.param_groups[0]['lr'])
        return lr



"""
hyperparameter_scans.py
=======================
Run randomized grid search to find ideal model hyperparameters, with possible deployments to batch system for scalability.
"""

import os, pandas as pd, numpy as np, subprocess
import time
from methylnet.torque_jobs import assemble_run_torque

def find_top_jobs(hyperparameter_input_csv,hyperparameter_output_log, n_top_jobs, crossover_p=0, val_loss_column='min_val_loss'):
    """Finds top performing jobs from hyper parameter scan to rerun and cross-over parameters.

    Parameters
    ----------
    hyperparameter_input_csv : str
        CSV file containing hyperparameter inputs.
    hyperparameter_output_log : str
        CSV file containing prior runs.
    n_top_jobs : int
        Number of top jobs to select
    crossover_p : float
        Rate of cross over of reused hyperparameters
    val_loss_column : str
        Loss column used to select top jobs

    Returns
    -------
    list
        List of list of new parameters for jobs to run.

    """

    custom_jobs = pd.read_csv(hyperparameter_input_csv)
    hyperparam_output = pd.read_csv(hyperparameter_output_log)[['job_name',val_loss_column]]
    best_outputs = hyperparam_output.sort_values(val_loss_column,ascending=True).iloc[:n_top_jobs,:]
    custom_jobs = custom_jobs[np.isin(custom_jobs['--job_name'].values,best_outputs['job_name'].values)]
    if custom_jobs.shape[0]==0 and best_outputs.shape[0]>0:
        custom_jobs = best_outputs.rename(columns={'--{}'.format(k):k for k in list(best_outputs)})[list(custom_jobs)]
        custom_jobs.loc[:,'--hidden_layer_topology']=custom_jobs.loc[:,'--hidden_layer_topology'].map(lambda x: x.replace('[','').replace(']',''))
    custom_jobs.loc[:,'--job_name']='False'
    if crossover_p:
        for j in range(1,custom_jobs.shape[1]):
            vals=custom_jobs.iloc[:,j].unique()
            for i in range(custom_jobs.shape[0]):
                if np.random.rand() <= crossover_p:
                    custom_jobs.iloc[i,j]=np.random.choice(vals)
    return [custom_jobs]

def replace_grid(old_grid, new_grid, topology_grid):
    """Replaces old hyperparameter search grid with one supplied by YAML."""
    #print(old_grid,new_grid)
    for k in new_grid.keys():
        if k != 'topology_grid':
            old_grid["--{}".format(k)] = new_grid[k]
    if 'topology_grid' in new_grid:
        topology_grid = new_grid.pop('topology_grid')
    return old_grid, topology_grid

def generate_topology(topology_grid, probability_decay_factor=0.9):
    """Generates list denoting neural network topology, list of hidden layer sizes.

    Parameters
    ----------
    topology_grid : list
        List of different hidden layer sizes (number neurons) to choose from.
    probability_decay_factor : float
        Degree of neural network model complexity for hyperparameter search. Search for less wide networks with a lower complexity value, bounded between 0 and infinity.

    Returns
    -------
    list
        List of hidden layer sizes.

    """

    probability_decay_factor=float(max(probability_decay_factor,0.))
    p = probability_decay_factor**np.arange(len(topology_grid))
    p /= sum(p)
    n_layer_attempts = int(round(3*probability_decay_factor))
    if n_layer_attempts:
        topology=list(filter(lambda x: x,np.random.choice(topology_grid,n_layer_attempts,p=p)))
    else:
        topology=[]
    if topology:
        return ','.join(map(str,topology))
    else:
        return ''
    return ''

def coarse_scan(hyperparameter_input_csv, hyperparameter_output_log, generate_input, job_chunk_size, stratify_column, reset_all, torque, gpu, gpu_node, nohup, mlp=False, custom_jobs=[], model_complexity_factor=0.9, set_beta=-1., n_jobs=4, categorical=True, add_softmax=False, additional_command = "", cuda=True, new_grid = {}):
    """Perform randomized hyperparameter grid search

    Parameters
    ----------
    hyperparameter_input_csv : type
        CSV file containing hyperparameter inputs.
    hyperparameter_output_log : type
        CSV file containing prior runs.
    generate_input : type
        Generate hyperparameter input csv.
    job_chunk_size : type
        Number of jobs to be launched at same time.
    stratify_column : type
        Performing classification?
    reset_all : type
        Rerun all jobs previously scanned.
    torque : type
        Run jobs using torque.
    gpu : type
        What GPU to use, set to -1 to be agnostic to GPU selection.
    gpu_node : type
        What GPU to use, set to -1 to be agnostic to GPU selection, for torque submission.
    nohup : type
        Launch jobs using nohup.
    mlp : type
        If running prediction job (classification/regression) after VAE.
    custom_jobs : type
        Supply custom job parameters to be run.
    model_complexity_factor : type
        Degree of neural network model complexity for hyperparameter search. Search for less wide networks with a lower complexity value, bounded between 0 and infinity.
    set_beta : type
        Don't hyperparameter scan over beta (KL divergence weight), and set it to value.
    n_jobs : type
        Number of jobs to generate.
    categorical : type
        Classification task?
    add_softmax : type
        Add softmax layer at end of neural network.
    cuda : type
        Whether to use GPU?
    """
    from itertools import cycle
    from pathos.multiprocessing import ProcessingPool as Pool
    os.makedirs(hyperparameter_input_csv[:hyperparameter_input_csv.rfind('/')],exist_ok=True)
    generated_input=[]
    np.random.seed(int(time.time()))
    if mlp:
        grid={'--learning_rate_vae':[1e-5,5e-5,1e-4,5e-4,1e-3,5e-3,1e-2,5e-2,1e-1],'--learning_rate_mlp':[1e-5,5e-5,1e-4,5e-4,1e-3,5e-3,1e-2,5e-2,1e-1,5e-1],
              '--weight_decay':[1e-4],'--n_epochs':[25,50,75,100,200,500,700], '--scheduler':['warm_restarts','null'], '--t_max':[10],
              '--eta_min':[1e-7,1e-6], '--t_mult':[1.,1.2,1.5,2],
              '--batch_size':[50,100,256,512], '--dropout_p':[0.,0.1,0.2,0.3,0.5],
              '--n_workers':[4], '--loss_reduction':['sum']}
        topology_grid = [0,100,200,300,500,1000,2000,3000,4096]
    else:
        grid={'--n_latent':[100,150,200,300,500],
              '--learning_rate':[5e-5,1e-4,5e-4,1e-3,5e-3,1e-2,5e-2,1e-1],
              '--weight_decay':[1e-4],'--n_epochs':[25,50,75,100,200,500,700],
              '--kl_warm_up':[0,20], '--beta':[0.,0.5,1,10,50,100,200,500] if set_beta == -1. else [set_beta],
              '--scheduler':['warm_restarts','null'], '--t_max':[10],
              '--eta_min':[1e-7,1e-6], '--t_mult':[1.,1.2,1.5,2],
              '--batch_size':[50,100,256,512],
              '--n_workers':[4], '--loss_reduction':['sum']}
        topology_grid=[0,100,200,300,500,1000,2000]
    if new_grid:
        grid,topology_grid=replace_grid(grid,new_grid,topology_grid)
    grid['--hidden_layer_topology' if mlp else '--hidden_layer_encoder_topology']=[generate_topology(topology_grid,probability_decay_factor=model_complexity_factor) for i in range(40)]
    if generate_input:
        for i in range(n_jobs):
            generated_input.append(['False']+[np.random.choice(grid[k]) for k in grid])
        generated_input=[pd.DataFrame(generated_input,columns=['--job_name']+list(grid.keys()))]
    if custom_jobs:
        custom_jobs[0].loc[:,'--job_name']='False'
        generated_input=custom_jobs
    def run(x):
        print(x)
        subprocess.call(x,shell=True)
    lower = lambda x: x.lower()
    if gpu == -1:
        gpus=cycle(range(4))
    else:
        gpus=cycle([gpu])
    if os.path.exists(hyperparameter_input_csv):
        df=pd.read_csv(hyperparameter_input_csv)
        df=[df[[col for col in list(df) if not col.startswith('Unnamed')]]]
    else:
        df = []
    df=pd.concat(df+generated_input,axis=0)[['--job_name']+list(grid.keys())].fillna('')
    print(df)
    if reset_all:
        df.loc[:,'--job_name']='False'
    df_final = df[df['--job_name'].astype(str).map(lower)=='false'].reset_index(drop=True)[list(grid.keys())]
    commands=[]
    for i in range(df_final.shape[0]):
        job_id = str(np.random.randint(0,100000000))
        if not mlp:
            commands.append('sh -c "methylnet-embed perform_embedding -bce {} -v -j {} -hl {} -sc {} {} && pymethyl-visualize transform_plot -i embeddings/vae_methyl_arr.pkl -o visualizations/{}_vae_embed.html -c {} -nn 10 "'.format("-c" if cuda else "",job_id,hyperparameter_output_log,stratify_column,' '.join(['{} {}'.format(k2,df_final.loc[i,k2]) for k2 in list(df_final) if (df_final.loc[i,k2] != '' and df_final.loc[i,k2] != np.nan)]),job_id,stratify_column))
        else:
            commands.append('sh -c "methylnet-predict make_prediction {} {} {} {} -v {} -j {} -hl {} {} && {}"'.format("-c" if cuda else "",'-sft' if add_softmax else '','-cat' if categorical else '',''.join([' -ic {}'.format(col) for col in stratify_column]),'-do' if stratify_column[0]=='disease_only' else '',job_id,hyperparameter_output_log,' '.join(['{} {}'.format(k2,df_final.loc[i,k2]) for k2 in list(df_final) if (df_final.loc[i,k2] != '' and df_final.loc[i,k2] != np.nan)]),
                                '&&'.join([" pymethyl-visualize transform_plot -i predictions/vae_mlp_methyl_arr.pkl -o visualizations/{}_{}_mlp_embed.html -c {} -nn 8 ".format(job_id,col,col) for col in stratify_column]))) #-do
        df.loc[np.arange(df.shape[0])==np.where(df['--job_name'].astype(str).map(lower)=='false')[0][0],'--job_name']=job_id
    for i in range(len(commands)):
        commands[i] = '{} {} {} {}'.format('CUDA_VISIBLE_DEVICES="{}"'.format(next(gpus)) if not torque else "",'nohup' if nohup else '',commands[i],'&' if nohup else '') # $gpuNum
    if torque:
        for command in commands:
            job = assemble_run_torque(command, use_gpu=cuda, additions=additional_command, queue='gpuq' if cuda else "normal", time=4, ngpu=1, additional_options='' if gpu_node == -1 else ' -l hostlist=g0{}'.format(gpu_node))
    else:
        if len(commands) == 1:
            subprocess.call(commands[0],shell=True)
        else:
            commands = np.array_split(commands,len(commands)//job_chunk_size)
            for command_list in commands:
                if nohup:
                    for command in command_list:
                        print(command)
                        subprocess.call(command,shell=True)
                else:
                    for command in command_list:
                        subprocess.call(command,shell=True)
                    """pool = Pool(len(command_list))
                    pool.map(run, command_list)
                    pool.close()
                    pool.join()"""
    df.to_csv(hyperparameter_input_csv)



"""
datasets.py
=======================
Contains datatypes core to loading MethylationArrays into Torch tensors.
"""

import torch
from torch import cat, stack, FloatTensor
from torch.utils.data import Dataset
from torchvision.transforms import Compose, RandomCrop, ToTensor, Scale, Pad
import pandas as pd, numpy as np
from sklearn.preprocessing import OneHotEncoder
from pymethylprocess.MethylationDataTypes import MethylationArray

def _reshape(cpgs_per_row, l):
    """Reshapes one sample's methylation values into image.

    Parameters
    ----------
    cpgs_per_row : int
        Number of CpGs per row
    l : int
        Number of splits.
    """
    def resize(vector):
        return vector[:,:int(l[1]/cpgs_per_row)*cpgs_per_row].reshape((l[0],int(l[1]/cpgs_per_row),cpgs_per_row))
    return resize

def convert_to_tensor():
    def to_tensor(arr):
        out= torch.FloatTensor(arr)
        return out
    return to_tensor

class Transformer:
    """Push methyl array sample to pytorch tensor, possibly turning into image.

    Parameters
    ----------
    convolutional : bool
        Whether running convolutions on torch dataset.
    cpg_per_row : int
        Number of CpGs per row of image if convolutional.
    l : tuple
        Attributes that describe image.

    Attributes
    ----------
    shape : tuple
        Shape of methylation array image for sample
    convolutional
    cpg_per_row
    l

    """
    def __init__(self, convolutional=False, cpg_per_row=30000, l=None):
        self.convolutional = convolutional
        self.cpg_per_row = cpg_per_row
        self.l = l
        if convolutional:
            self.shape=(self.l[0],int(self.l[1]/cpgs_per_row),cpgs_per_row)
        else:
            self.shape=None

    def generate(self):
        """Generate function for transform.

        Returns
        -------
        function

        """
        data_aug = []
        if self.convolutional and self.l != None:
            data_aug.append(_reshape(self.cpgs_per_row, self.l))
        data_aug.append(convert_to_tensor())
        return Compose(data_aug)

# From Titus
# test_set_percent = 0.1
# methyl_test_df = methyl_df2.sample(frac=test_set_percent)
# methyl_train_df = methyl_df2.drop(methyl_test_df.index)
# instead get beta keys and sample

def get_methylation_dataset(methylation_array, outcome_col, convolutional=False, cpg_per_row=1200, predict=False, categorical=False, categorical_encoder=False, generate=False):
    """Turn methylation array into pytorch dataset.

    Parameters
    ----------
    methylation_array : MethylationArray
        Input MethylationArray.
    outcome_col : str
        Pheno column to train on.
    convolutional : bool
        Whether running CNN on methylation data
    cpg_per_row : int
        If convolutional, number of cpgs per image row.
    predict : bool
        Running prediction algorithm vs VAE.
    categorical : bool
        Whether training on categorical vs continuous variables.
    categorical_encoder :
        Scikit learn encoder.

    Returns
    -------
    Pytorch Dataset

    """
    if generate:
        return MethylationGenerationDataSet(methylation_array, Transformer(convolutional, cpg_per_row, methylation_array.beta.shape), outcome_col, categorical=categorical, categorical_encoder=categorical_encoder)
    elif predict:
        return MethylationPredictionDataSet(methylation_array, Transformer(convolutional, cpg_per_row, methylation_array.beta.shape), outcome_col, categorical=categorical, categorical_encoder=categorical_encoder)
    else:
        return MethylationDataSet(methylation_array, Transformer(convolutional, cpg_per_row, methylation_array.beta.shape), outcome_col)

class MethylationDataSet(Dataset):
    """Pytorch Dataset that contains instances of methylation array samples to be loaded.

    Parameters
    ----------
    methylation_array : MethylationArray
        Methylation Array input.
    transform : Transformer
        Transforms data into torch tensor.
    outcome_col : str
        Pheno column(s) to train on.
    categorical : bool
        Whether predicting categorical/classification.
    categorical_encoder : encoder
        Encoder used to binarize categorical column.

    Attributes
    ----------
    samples : np.array
        Samples of MethylationArray
    features : np.array
        List CpGs.
    encoder :
        Encoder used to binarize categorical column.
    new_shape :
        Shape of torch tensors.
    length :
        Number CpGs.
    methylation_array
    outcome_col
    transform

    """
    def __init__(self, methylation_array, transform, outcome_col='', categorical=False, categorical_encoder=False):
        self.methylation_array = methylation_array
        self.outcome_col = self.methylation_array.pheno.loc[:,outcome_col] if outcome_col else pd.Series(np.ones(len(self)),index=self.methylation_array.pheno.index)
        self.outcome_col = self.outcome_col.loc[self.methylation_array.beta.index,]
        self.samples = np.array(list(self.methylation_array.beta.index))
        self.features = np.array(list(self.methylation_array.beta))
        self.methylation_array.beta = self.methylation_array.beta.values
        self.samples = np.array(list(self.outcome_col.index))
        self.outcome_col=self.outcome_col.values
        if categorical:
            if len(self.outcome_col.shape)<2 or self.outcome_col.shape[0]==1:
                self.outcome_col=self.outcome_col[:,np.newaxis]
            #self.outcome_col=self.outcome_col#[:,np.newaxis]
            if not categorical_encoder:
                print(self.outcome_col)
                self.encoder = OneHotEncoder(sparse=False)
                self.encoder.fit(self.outcome_col)
            else:
                self.encoder=categorical_encoder
            self.outcome_col=self.encoder.transform(self.outcome_col)
        else:
            self.encoder = None
        self.transform = transform
        self.new_shape = self.transform.shape
        self.length = self.methylation_array.beta.shape[0]
        #print(self.outcome_col)
        print(self.outcome_col.shape)

    def to_methyl_array(self):
        """Convert torch dataset back into methylation array, useful because turning into torch dataset can cause the original MethylationArray beta matrix to turn into numpy array, when needs turn back into pandas dataframe.

        Returns
        -------
        MethylationArray

        """
        return MethylationArray(self.methylation_array.pheno,pd.DataFrame(self.methylation_array.beta,index=self.samples,columns=self.features),'')

    def __getitem__(self,index):
        return self.transform.generate()(self.methylation_array.beta[index,:]),self.samples[index],self.outcome_col[index]
        """if self.mlp:
            return self.get_item_mlp(index)
        else:
            return self.get_item_vae(index)"""

    """def get_item_vae(self, index):# .iloc[index,] [index] .iloc[index]
        transform = self.transform.generate()
        #print(self.methylation_array.beta.values.shape)
        return transform(self.methylation_array.beta.values),self.samples.tolist(),self.outcome_col.values.tolist()

    def get_item_mlp(self, index):# .iloc[index,] [index] .iloc[index]
        transform = self.transform.generate()
        #print(self.methylation_array.beta.values.shape)
        return transform(self.methylation_array.beta.iloc[index,:].values),self.samples[index],self.outcome_col.iloc[index,:].values
        """
    def __len__(self):
        return self.length

class MethylationPredictionDataSet(MethylationDataSet):
    """MethylationArray Torch Dataset that contains instances of methylation array samples to be loaded, specifically targetted for prediction.

    Parameters
    ----------
    methylation_array : MethylationArray
        Methylation Array input.
    transform : Transformer
        Transforms data into torch tensor.
    outcome_col : str
        Pheno column(s) to train on.
    categorical : bool
        Whether predicting categorical/classification.
    categorical_encoder : encoder
        Encoder used to binarize categorical column.

    Attributes
    ----------
    samples : np.array
        Samples of MethylationArray
    features : np.array
        List CpGs.
    encoder :
        Encoder used to binarize categorical column.
    new_shape :
        Shape of torch tensors.
    length :
        Number CpGs.
    methylation_array
    outcome_col
    transform

    """
    def __init__(self, methylation_array, transform, outcome_col='', categorical=False, categorical_encoder=False):
        super().__init__(methylation_array, transform, outcome_col, categorical, categorical_encoder=categorical_encoder)
        #print(self.outcome_col)

    def __getitem__(self,index):
        return self.transform.generate()(self.methylation_array.beta[index,:]),self.samples[index],torch.FloatTensor(self.outcome_col[index,:])

class MethylationGenerationDataSet(MethylationPredictionDataSet):
    """MethylationArray Torch Dataset that contains instances of methylation array samples to be loaded, specifically targetted for prediction.

    Parameters
    ----------
    methylation_array : MethylationArray
        Methylation Array input.
    transform : Transformer
        Transforms data into torch tensor.
    outcome_col : str
        Pheno column(s) to train on.
    categorical : bool
        Whether predicting categorical/classification.
    categorical_encoder : encoder
        Encoder used to binarize categorical column.

    Attributes
    ----------
    samples : np.array
        Samples of MethylationArray
    features : np.array
        List CpGs.
    encoder :
        Encoder used to binarize categorical column.
    new_shape :
        Shape of torch tensors.
    length :
        Number CpGs.
    methylation_array
    outcome_col
    transform

    """
    def __init__(self, methylation_array, transform, outcome_col='', categorical=False, categorical_encoder=False):
        super().__init__(methylation_array, transform, outcome_col, categorical, categorical_encoder=categorical_encoder)

    def __getitem__(self,index):
        #print(index)
        y=self.outcome_col[index]
        if not np.shape(y):
            y=np.array([y])
        return self.transform.generate()(self.methylation_array.beta[index,:]),torch.FloatTensor(y)

class RawBetaArrayDataSet(Dataset):
    """Torch Dataset just for beta matrix in numpy format.

    Parameters
    ----------
    beta_array : numpy.array
        Beta value matrix in numpy form.
    transform :
        Transforms data to torch tensor.

    Attributes
    ----------
    length :
        Number CpGs.
    beta_array
    transform

    """
    def __init__(self, beta_array, transform):
        self.beta_array = beta_array
        self.transform = transform
        self.length = self.beta_array.shape[0]

    def __getitem__(self,index):
        return self.transform.generate()(self.beta_array[index,:])

    def __len__(self):
        return self.length



import click, os
from methylnet.torque_jobs import *

CONTEXT_SETTINGS = dict(help_option_names=['-h','--help'], max_content_width=90)

@click.group(context_settings= CONTEXT_SETTINGS)
@click.version_option(version='0.1')
def torque():
    pass

@torque.command()
@click.option('-c', '--command', default='', help='Command to execute through torque.', show_default=True)
@click.option('-gpu', '--use_gpu', is_flag=True, help='Specify whether to use GPUs.', show_default=True)
@click.option('-a', '--additions', default='', help='Additional commands to add, one liner for now.', show_default=True)
@click.option('-q', '--queue', default='default', help='Queue for torque submission, gpuq also a good one if using GPUs.', show_default=True)
@click.option('-t', '--time', default=1, help='Walltime in hours for job.', show_default=True)
@click.option('-n', '--ngpu', default=0, help='Number of gpus to request.', show_default=True)
def run_torque_job(command, use_gpu, additions, queue, time, ngpu):
    """Run torque job."""
    replace_dict = assemble_replace_dict(command, use_gpu, additions, queue, time, ngpu)
    run_torque_job_(replace_dict)

if __name__ == '__main__':
    torque()



from pymethylprocess.MethylationDataTypes import MethylationArray,MethylationArrays, extract_pheno_beta_df_from_pickle_dict
import pickle
import pandas as pd, numpy as np
import click
import os, subprocess
from os.path import join
import time

CONTEXT_SETTINGS = dict(help_option_names=['-h','--help'], max_content_width=90)

@click.group(context_settings= CONTEXT_SETTINGS)
@click.version_option(version='0.1')
def embed():
    pass

def embed_vae(train_pkl,output_dir,cuda,n_latent,lr,weight_decay,n_epochs,hidden_layer_encoder_topology, kl_warm_up=0, beta=1., scheduler='null', decay=0.5, t_max=10, eta_min=1e-6, t_mult=2, bce_loss=False, batch_size=50, val_pkl='val_methyl_array.pkl', n_workers=9, convolutional = False, height_kernel_sizes=[], width_kernel_sizes=[], add_validation_set=False, loss_reduction='sum', stratify_column='disease'):
    from methylnet.models import AutoEncoder, TybaltTitusVAE
    from methylnet.datasets import get_methylation_dataset
    import torch
    from torch.utils.data import DataLoader
    from torch.nn import MSELoss, BCELoss
    os.makedirs(output_dir,exist_ok=True)

    output_file = join(output_dir,'output_latent.csv')
    output_model = join(output_dir,'output_model.p')
    training_curve_file = join(output_dir, 'training_val_curve.p')
    outcome_dict_file = join(output_dir,'output_outcomes.p')
    output_pkl = join(output_dir, 'vae_methyl_arr.pkl')

    #input_dict = pickle.load(open(input_pkl,'rb'))
    #methyl_array=MethylationArray(*extract_pheno_beta_df_from_pickle_dict(input_dict))
    #print(methyl_array.beta)
    train_methyl_array, val_methyl_array = MethylationArray.from_pickle(train_pkl), MethylationArray.from_pickle(val_pkl)#methyl_array.split_train_test(train_p=train_percent, stratified=True, disease_only=True, key='disease', subtype_delimiter=',')

    train_methyl_dataset = get_methylation_dataset(train_methyl_array,stratify_column) # train, test split? Add val set?

    val_methyl_dataset = get_methylation_dataset(val_methyl_array,stratify_column)

    if not batch_size:
        batch_size=len(methyl_dataset)

    train_batch_size = min(batch_size,len(train_methyl_dataset))
    val_batch_size = min(batch_size,len(val_methyl_dataset))

    train_methyl_dataloader = DataLoader(
        dataset=train_methyl_dataset,
        num_workers=n_workers,#n_workers
        batch_size=train_batch_size,
        shuffle=True,
        pin_memory=False)

    val_methyl_dataloader = DataLoader(
        dataset=val_methyl_dataset,
        num_workers=n_workers,
        batch_size=val_batch_size,
        shuffle=True,
        pin_memory=False)

    scaling_factors = dict(train=float(len(train_methyl_dataset))/((len(train_methyl_dataset)//train_batch_size)*train_batch_size),
                           val=float(len(val_methyl_dataset))/((len(val_methyl_dataset)//val_batch_size)*val_batch_size),
                           train_batch_size=train_batch_size,val_batch_size=val_batch_size)
    print('SCALE',len(train_methyl_dataset),len(val_methyl_dataset),train_batch_size,val_batch_size, scaling_factors)
    n_input = train_methyl_array.return_shape()[1]
    if not convolutional:
        model=TybaltTitusVAE(n_input=n_input,n_latent=n_latent,hidden_layer_encoder_topology=hidden_layer_encoder_topology,cuda=cuda)
    else:
        model = CVAE(n_latent=n_latent,in_shape=methyl_dataset.new_shape, kernel_heights=height_kernel_sizes, kernel_widths=width_kernel_sizes, n_pre_latent=n_latent*2) # change soon

    optimizer = torch.optim.Adam(model.parameters(), lr = lr, weight_decay=weight_decay)
    loss_fn = BCELoss(reduction=loss_reduction) if bce_loss else MSELoss(reduction=loss_reduction) # 'sum'
    scheduler_opts=dict(scheduler=scheduler,lr_scheduler_decay=decay,T_max=t_max,eta_min=eta_min,T_mult=t_mult)
    auto_encoder=AutoEncoder(autoencoder_model=model,n_epochs=n_epochs,loss_fn=loss_fn,optimizer=optimizer,cuda=cuda,kl_warm_up=kl_warm_up,beta=beta, scheduler_opts=scheduler_opts)
    if add_validation_set:
        auto_encoder.add_validation_set(val_methyl_dataloader)
    auto_encoder = auto_encoder.fit(train_methyl_dataloader)
    train_methyl_array=train_methyl_dataset.to_methyl_array()
    val_methyl_array=val_methyl_dataset.to_methyl_array()
    del val_methyl_dataloader, train_methyl_dataloader, val_methyl_dataset, train_methyl_dataset

    methyl_dataset=get_methylation_dataset(MethylationArrays([train_methyl_array,val_methyl_array]).combine(),stratify_column)
    methyl_dataset_loader = DataLoader(
        dataset=methyl_dataset,
        num_workers=n_workers,
        batch_size=1,
        shuffle=False)
    latent_projection, sample_names, _ = auto_encoder.transform(methyl_dataset_loader)
    #print(latent_projection.shape)
    methyl_array = methyl_dataset.to_methyl_array()
    #sample_names = np.array([sample_name[0] for sample_name in sample_names]) # FIXME
    #outcomes = np.array([outcome[0] for outcome in outcomes]) # FIXME
    #outcome_dict=dict(zip(sample_names,outcomes))
    #print(methyl_array.beta)
    latent_projection=pd.DataFrame(latent_projection,index=methyl_array.beta.index)
    methyl_array.beta=latent_projection
    methyl_array.write_pickle(output_pkl)
    latent_projection.to_csv(output_file)
    pickle.dump(auto_encoder.training_plot_data,open(training_curve_file,'wb'))
    torch.save(auto_encoder.model,output_model)
    #pickle.dump(outcome_dict, open(outcome_dict_file,'wb'))
    return latent_projection, None, scaling_factors, n_input, auto_encoder

@embed.command()
@click.option('-i', '--train_pkl', default='./train_val_test_sets/train_methyl_array.pkl', help='Input database for beta and phenotype data.', type=click.Path(exists=False), show_default=True)
@click.option('-o', '--output_dir', default='./embeddings/', help='Output directory for embeddings.', type=click.Path(exists=False), show_default=True)
@click.option('-c', '--cuda', is_flag=True, help='Use GPUs.')
@click.option('-n', '--n_latent', default=64, help='Number of latent dimensions.', show_default=True)
@click.option('-lr', '--learning_rate', default=1e-3, help='Learning rate.', show_default=True)
@click.option('-wd', '--weight_decay', default=1e-4, help='Weight decay of adam optimizer.', show_default=True)
@click.option('-e', '--n_epochs', default=50, help='Number of epochs to train over.', show_default=True)
@click.option('-hlt', '--hidden_layer_encoder_topology', default='', help='Topology of hidden layers, comma delimited, leave empty for one layer encoder, eg. 100,100 is example of 5-hidden layer topology.', type=click.Path(exists=False), show_default=True)
@click.option('-kl', '--kl_warm_up', default=0, help='Number of epochs before introducing kl_loss.', show_default=True)
@click.option('-b', '--beta', default=1., help='Weighting of kl divergence.', show_default=True)
@click.option('-s', '--scheduler', default='null', help='Type of learning rate scheduler.', type=click.Choice(['null','exp','warm_restarts']),show_default=True)
@click.option('-d', '--decay', default=0.5, help='Learning rate scheduler decay for exp selection.', show_default=True)
@click.option('-t', '--t_max', default=10, help='Number of epochs before cosine learning rate restart.', show_default=True)
@click.option('-eta', '--eta_min', default=1e-6, help='Minimum cosine LR.', show_default=True)
@click.option('-m', '--t_mult', default=2., help='Multiply current restart period times this number given number of restarts.', show_default=True)
@click.option('-bce', '--bce_loss', is_flag=True, help='Use bce loss instead of MSE.')
@click.option('-bs', '--batch_size', default=50, show_default=True, help='Batch size.')
@click.option('-vp', '--val_pkl', default='./train_val_test_sets/val_methyl_array.pkl', help='Validation Set Methylation Array Location.', show_default=True, type=click.Path(exists=False),)
@click.option('-w', '--n_workers', default=9, show_default=True, help='Number of workers.')
@click.option('-conv', '--convolutional', is_flag=True, help='Use convolutional VAE.')
@click.option('-hs', '--height_kernel_sizes', default=[], multiple=True, help='Heights of convolutional kernels.')
@click.option('-ws', '--width_kernel_sizes', default=[], multiple=True, help='Widths of convolutional kernels.')
@click.option('-v', '--add_validation_set', is_flag=True, help='Evaluate validation set.')
@click.option('-l', '--loss_reduction', default='sum', show_default=True, help='Type of reduction on loss function.', type=click.Choice(['sum','elementwise_mean','none']))
@click.option('-hl', '--hyperparameter_log', default='embeddings/embed_hyperparameters_log.csv', show_default=True, help='CSV file containing prior runs.', type=click.Path(exists=False))
@click.option('-sc', '--stratify_column', default='disease', show_default=True, help='Column to stratify samples on.', type=click.Path(exists=False))
@click.option('-j', '--job_name', default='embed_job', show_default=True, help='Embedding job name.', type=click.Path(exists=False))
def perform_embedding(train_pkl,output_dir,cuda,n_latent,learning_rate,weight_decay,n_epochs,hidden_layer_encoder_topology, kl_warm_up, beta, scheduler, decay, t_max, eta_min, t_mult, bce_loss, batch_size, val_pkl, n_workers, convolutional, height_kernel_sizes, width_kernel_sizes, add_validation_set, loss_reduction, hyperparameter_log,stratify_column,job_name):
    """Perform variational autoencoding on methylation dataset."""
    hlt_list=filter(None,hidden_layer_encoder_topology.split(','))
    if hlt_list:
        hidden_layer_encoder_topology=list(map(int,hlt_list))
    else:
        hidden_layer_encoder_topology=[]
    _,_,scaling_factors,n_input,autoencoder = embed_vae(train_pkl,output_dir,cuda,n_latent,learning_rate,weight_decay,n_epochs,hidden_layer_encoder_topology,kl_warm_up,beta, scheduler, decay, t_max, eta_min, t_mult, bce_loss, batch_size, val_pkl, n_workers, convolutional, height_kernel_sizes, width_kernel_sizes, add_validation_set, loss_reduction, stratify_column)
    hyperparameter_row = [job_name,n_epochs, autoencoder.best_epoch, autoencoder.min_loss, autoencoder.min_val_loss, autoencoder.min_val_kl_loss, autoencoder.min_val_recon_loss, autoencoder.min_loss*scaling_factors['train'], autoencoder.min_val_loss*scaling_factors['val'], autoencoder.min_val_kl_loss*scaling_factors['val'], autoencoder.min_val_recon_loss*scaling_factors['val'], autoencoder.min_val_recon_loss+(autoencoder.min_val_kl_loss/beta if beta > 0. else 0.), n_input, n_latent, str(hidden_layer_encoder_topology), learning_rate, weight_decay, beta, kl_warm_up, scheduler, t_max, t_mult, scaling_factors['train_batch_size'], scaling_factors['val_batch_size']]
    hyperparameter_df = pd.DataFrame(columns=['job_name','n_epochs',"best_epoch", "min_loss", "min_val_loss", "min_val_kl_loss", "min_val_recon_loss", "min_loss-batchsize_adj", "min_val_loss-batchsize_adj", "min_val_kl_loss-batchsize_adj", "min_val_recon_loss-batchsize_adj", "min_val_beta-adj_loss", "n_input", "n_latent", "hidden_layer_encoder_topology", "learning_rate", "weight_decay", "beta", "kl_warm_up", "scheduler", "t_max", "t_mult", "train_batch_size",'val_batch_size'])
    hyperparameter_df.loc[0] = hyperparameter_row
    if os.path.exists(hyperparameter_log):
        print('APPEND')
        hyperparameter_df_former = pd.read_csv(hyperparameter_log)
        hyperparameter_df_former=hyperparameter_df_former[[col for col in list(hyperparameter_df) if not col.startswith('Unnamed')]]
        hyperparameter_df=pd.concat([hyperparameter_df_former,hyperparameter_df],axis=0)
    hyperparameter_df.to_csv(hyperparameter_log)

@embed.command()
@click.option('-hcsv', '--hyperparameter_input_csv', default='embeddings/embed_hyperparameters_scan_input.csv', show_default=True, help='CSV file containing hyperparameter inputs.', type=click.Path(exists=False))
@click.option('-hl', '--hyperparameter_output_log', default='embeddings/embed_hyperparameters_log.csv', show_default=True, help='CSV file containing prior runs.', type=click.Path(exists=False))
@click.option('-g', '--generate_input', is_flag=True, help='Generate hyperparameter input csv.')
@click.option('-c', '--job_chunk_size', default=4, help='If not series, chunk up and run these number of commands at once..')
@click.option('-sc', '--stratify_column', default='disease', show_default=True, help='Column to stratify samples on.', type=click.Path(exists=False))
@click.option('-r', '--reset_all', is_flag=True, help='Run all jobs again.')
@click.option('-t', '--torque', is_flag=True, help='Submit jobs on torque.')
@click.option('-gpu', '--gpu', default=-1, help='If torque submit, which gpu to use.', show_default=True)
@click.option('-gn', '--gpu_node', default=-1, help='If torque submit, which gpu node to use.', show_default=True)
@click.option('-nh', '--nohup', is_flag=True, help='Nohup launch jobs.')
@click.option('-mc', '--model_complexity_factor', default=1., help='Degree of neural network model complexity for hyperparameter search. Search for less wide and less deep networks with a lower complexity value, bounded between 0 and infinity.', show_default=True)
@click.option('-b', '--set_beta', default=1., help='Set beta value, bounded between 0 and infinity. Set to -1 ', show_default=True)
@click.option('-j', '--n_jobs', default=4, help='Number of jobs to generate.')
@click.option('-n', '--n_jobs_relaunch', default=0, help='Relaunch n top jobs from previous run.', show_default=True)
@click.option('-cp', '--crossover_p', default=0., help='Rate of crossover between hyperparameters.', show_default=True)
@click.option('-v', '--val_loss_column', default='min_val_loss-batchsize_adj', help='Validation loss column.', type=click.Path(exists=False))
@click.option('-a', '--additional_command', default='', help='Additional command to input for torque run.', type=click.Path(exists=False))
@click.option('-cu', '--cuda', is_flag=True, help='Use GPUs.')
@click.option('-grid', '--hyperparameter_yaml', default='', help='YAML file with custom subset of hyperparameter grid.', type=click.Path(exists=False))
def launch_hyperparameter_scan(hyperparameter_input_csv, hyperparameter_output_log, generate_input, job_chunk_size, stratify_column, reset_all, torque, gpu, gpu_node, nohup, model_complexity_factor, set_beta,n_jobs, n_jobs_relaunch, crossover_p, val_loss_column, additional_command, cuda, hyperparameter_yaml):
    """Launch randomized grid search of neural network hyperparameters."""
    from methylnet.hyperparameter_scans import coarse_scan, find_top_jobs
    custom_jobs=[]
    if n_jobs_relaunch:
        custom_jobs=find_top_jobs(hyperparameter_input_csv, hyperparameter_output_log,n_jobs_relaunch, crossover_p, val_loss_column)
    if os.path.exists(hyperparameter_yaml):
        from ruamel.yaml import safe_load as load
        with open(hyperparameter_yaml) as f:
            new_grid = load(f)
        #print(new_grid)
    else:
        new_grid = {}
    coarse_scan(hyperparameter_input_csv, hyperparameter_output_log, generate_input, job_chunk_size, stratify_column, reset_all, torque, gpu, gpu_node, nohup, mlp=False, model_complexity_factor=model_complexity_factor, set_beta=set_beta,n_jobs=n_jobs, custom_jobs=custom_jobs, additional_command=additional_command, cuda=cuda, new_grid=new_grid)

#################

if __name__ == '__main__':
    embed()
